# built-in imports
from pathlib import Path
from typing import Union, Tuple, List, Optional, Type

# third-party import
import lief

from libpastis import FuzzingEngineDescriptor, EngineConfiguration
from libpastis.types import ExecMode, CoverageMode, FuzzMode


class AFLConfigurationInterface(EngineConfiguration):
    """
    Small wrapping function for AFL++ additional parameters
    """

    def __init__(self, args: List[str] = None):
        self._argvs = [] if args is None else args # Argument to send on the command line

    @staticmethod
    def new() -> 'AFLConfigurationInterface':
        return AFLConfigurationInterface()

    @staticmethod
    def from_file(filepath: Path) -> 'AFLConfigurationInterface':
        with open(filepath, "r") as f:
            return AFLConfigurationInterface(f.read().split())

    @staticmethod
    def from_str(s: str) -> 'AFLConfigurationInterface':
        return AFLConfigurationInterface(s.split())

    def to_str(self) -> str:
        return " ".join(self._argvs)

    def get_coverage_mode(self) -> CoverageMode:
        """ Current coverage mode selected in the file """
        return CoverageMode.AUTO

    def set_target(self, target: int) -> None:
        # Note: Giving a target to Honggfuzz does not
        # do anything as Honggfuzz is not directed.
        pass


class AFLPPEngineDescriptor(FuzzingEngineDescriptor):

    NAME = "AFLPP"
    SHORT_NAME = "AFLPP"
    VERSION = "1.0.0"  # Should be in sync with alfpp.__version__

    config_class = AFLConfigurationInterface

    def __init__(self):
        pass

    @staticmethod
    def accept_file(binary_file: Path) -> Tuple[bool, Optional[ExecMode], Optional[FuzzMode]]:
        if str(binary_file).endswith(".cmplog"):
            return False, None, None

        p = lief.parse(str(binary_file))
        if not p:
            return False, None, None

        # Search for HF instrumentation
        instrumented = False

        for s in p.symbols:
            if "__afl_" in s.name:
                instrumented = True
                break

        for f in p.functions:
            if "__afl_" in f.name:
                instrumented = True
                break

        if not instrumented:
            # NOTE This can be improve. We usually use PERSISTENT mode when
            # fuzzing a binary-only target because of performance reasons but
            # it can also be SINGLE_EXEC. Therefore, ExecMode would not be the
            # right place to add the BINARY_ONLY option (it was done this way
            # to keep things simple).
            return True, ExecMode.AUTO, FuzzMode.BINARY_ONLY

        return True, ExecMode.AUTO, FuzzMode.INSTRUMENTED

    @staticmethod
    def supported_coverage_strategies() -> List[CoverageMode]:
        return [CoverageMode.AUTO]



# builtin imports
from pathlib import Path
from typing import List, Tuple, Optional

# Local imports
from libpastis.types import CoverageMode, ExecMode, FuzzMode


class EngineConfiguration(object):
    """
    Basic interface to represent an engine configuration file
    on broker side. A fuzzing engine have to provide such object
    so that the broker can load them and forwarding them to clients.
    """

    @staticmethod
    def new() -> 'EngineConfiguration':
        """
        Static method that should return a fresh configuration object.

        :return: Configuration object
        """
        raise NotImplementedError

    @staticmethod
    def from_file(filepath: Path) -> 'EngineConfiguration':
        """
        Load a configuration object from file.

        :param filepath: Path to the configuration
        :return: Configuration object
        """
        raise NotImplementedError

    @staticmethod
    def from_str(s: str) -> 'EngineConfiguration':
        """
        Parse a string to a configuration object.

        :param s: configuration as string
        :return: configuration object
        """
        raise NotImplementedError

    def to_str(self) -> str:
        """
        Serialize configuration object to string.

        :return: serialize configuration
        """
        raise NotImplementedError

    def get_coverage_mode(self) -> CoverageMode:
        """
        Should return the coverage mode defined in the configuration.
        For greybox fuzzer like AFL++, Honggfuzz one can return :py:obj:`CoverageMode.AUTO`.
        If the engine support different coverage metric it should return
        the one selected.

        :return: coverage mode used
        """
        raise NotImplementedError

    def set_target(self, target: int) -> None:
        """
        Set a specific target (address, index etc), that should be targeted by
        the fuzzing engine. This will be used when running in a targeted way.

        :param target: identifier of the target
        """
        pass


class FuzzingEngineDescriptor(object):
    """
    Abstract class describing a fuzzer engine. This object is used on
    broker side, to identify the name and version of a fuzzer and to
    know whether or not it accept a specific executable file.
    """

    NAME = "abstract-engine"
    #: Name of the fuzzing Engine
    SHORT_NAME = "AE"
    #: Short name of the fuzzing engine
    VERSION = "1.0"
    #: Version of the engine

    config_class = EngineConfiguration
    #: Configuration class associated with the engine

    @staticmethod
    def accept_file(binary_file: Path) -> Tuple[bool, Optional[ExecMode], Optional[FuzzMode]]:
        """
        Function called by the broker with all executable files detected in its directory.
        As an fuzzer developer, you have to implement this function to indicate whether
        a file is accepted as a target or not.

        :param binary_file: file path to an executable file
        :return: True if supported, and two optional attributes indicating the ExecMode and FuzzMode
        """
        raise NotImplementedError()

    @staticmethod
    def supported_coverage_strategies() -> List[CoverageMode]:
        """
        List of coverage metrics supported by the fuzzer. If it only
        support a single one, it should be :py:obj:`CoverageMode.AUTO`.

        :return: list of coverage modes
        """
        raise NotImplementedError()



# built-in imports
from pathlib import Path
from typing import Union, Tuple, List, Optional
import json

# third-party import
import lief
from typing import Type

from libpastis import FuzzingEngineDescriptor, EngineConfiguration
from libpastis.types import ExecMode, CoverageMode, FuzzMode


class HonggfuzzConfigurationInterface(EngineConfiguration):
    def __init__(self, args: List[str] = None):
        self._argvs = [] if args is None else args # Argument to send on the command line

    @staticmethod
    def new() -> 'HonggfuzzConfigurationInterface':
        return HonggfuzzConfigurationInterface()

    @staticmethod
    def from_file(filepath: Path) -> 'HonggfuzzConfigurationInterface':
        with open(filepath, "r") as f:
            return HonggfuzzConfigurationInterface(f.read().split())

    @staticmethod
    def from_str(s: str) -> 'HonggfuzzConfigurationInterface':
        return HonggfuzzConfigurationInterface(s.split())

    def to_str(self) -> str:
        return " ".join(self._argvs)

    def get_coverage_mode(self) -> CoverageMode:
        """
        Current coverage mode selected in the file.
        Always EDGE for Honggfuzz
        """
        return CoverageMode.AUTO

    def set_target(self, target: int) -> None:
        # Note: Giving a target to Honggfuzz does not
        # do anything as Honggfuzz is not directed.
        pass


class HonggfuzzEngineDescriptor(FuzzingEngineDescriptor):

    NAME = "HONGGFUZZ"
    SHORT_NAME = "HF"
    VERSION = "1.0.0"  # Should be in sync with hfwrapper.__version__

    HF_PERSISTENT_SIG = b"\x01_LIBHFUZZ_PERSISTENT_BINARY_SIGNATURE_\x02\xFF"

    config_class = HonggfuzzConfigurationInterface

    def __init__(self):
        pass

    @staticmethod
    def accept_file(binary_file: Path) -> Tuple[bool, Optional[ExecMode], Optional[FuzzMode]]:
        p = lief.parse(str(binary_file))
        if not p:
            return False, None, None

        # Search for HF instrumentation
        instrumented = False
        for f in p.functions:
            if "hfuzz_" in f.name:
                instrumented = True
                break
        if not instrumented:
            return True, ExecMode.PERSISTENT, FuzzMode.BINARY_ONLY

        # Search for persistent magic
        exmode = ExecMode.SINGLE_EXEC  # by default single_exec
        sections = {x.name: x for x in p.sections}
        if '.rodata' in sections:
            rodata_content = bytearray(sections['.rodata'].content)
            if HonggfuzzEngineDescriptor.HF_PERSISTENT_SIG in rodata_content:
                exmode = ExecMode.PERSISTENT
        else:
            if 'HF_ITER' in (x.name for x in p.imported_functions):  # More dummy method
                exmode = ExecMode.PERSISTENT
        return True, exmode, FuzzMode.INSTRUMENTED

    @staticmethod
    def supported_coverage_strategies() -> List[CoverageMode]:
        return [CoverageMode.AUTO]



# built-in imports
from pathlib import Path
import zipfile
import tempfile
import logging
from typing import Tuple, Optional, Union

# third-party imports
import lief
import magic
import shutil
import stat

# local imports
from libpastis.types import Arch, Platform


class BinaryPackage(object):
    """
    Binary Package representing a given target to fuzz along with its shared
    libraries and additional files required (cmplog, dictionnary etc.).
    This object is received by fuzzing agents as part of the START message.
    """

    EXTENSION_BLACKLIST = ['.gt', '.Quokka', '.quokka', '.cmplog']
    #: specific extensions that will be ignored for the `other_files`

    def __init__(self, main_binary: Path):
        """
        :param main_binary: main executable file path
        """
        self._main_bin = Path(main_binary)
        self._quokka = None
        self._callgraph = None
        self._cmplog = None
        self._dictionary = None
        self.other_files = []
        #: list of additional files contained in this package

        self._package_file = None
        self._arch = None
        self._platform = None

    @property
    def executable_path(self) -> Path:
        """
        Path to the main executable file to fuzz.

        :return: filepath
        """
        return self._main_bin

    @property
    def name(self) -> str:
        """
        Name of the executable file

        :return: name as a string
        """
        return self._main_bin.name

    @property
    def quokka(self) -> Optional[Path]:
        """
        Path to the quokka file if provided.

        :return: path of the quokka file
        """
        return self._quokka

    @property
    def callgraph(self) -> Optional[Path]:
        """
        Path to the callgraph file if provided.

        :return: path of the quokka file
        """
        return self._callgraph

    @property
    def cmplog(self) -> Optional[Path]:
        """
        Path to the complog executable file if provided.

        :return: path to the complog file
        """
        return self._cmplog

    @property
    def dictionary(self) -> Optional[Path]:
        """
        Path the to dictionnary file if provided.

        :return: path to the dictionnary file
        """
        return self._dictionary

    def is_cmplog(self) -> bool:
        """
        Check if the package contains a cmplog file.

        :return: True if contains cmplog
        """
        return self._cmplog is not None

    def is_quokka(self) -> bool:
        """
        Check if the package contains a quokka file.

        :return: True if contains a quokka file
        """
        return self._quokka is not None

    def is_dictionary(self) -> bool:
        """
        Check if the package contains a dictionnary.

        :return: True if contains a dictionnary
        """
        return self._dictionary is not None

    def is_standalone(self) -> bool:
        """
        Indicates that this BinaryPackage only contains the program under test and no
        additional files such as a Quokka database or a cmplog instrumented binary.
        This is used in pastis-broker when sending the 'start' command to agents.
        """
        return not (self.is_quokka() or
                    self.is_cmplog() or
                    self.is_dictionary() or
                    bool(self.other_files))

    @property
    def arch(self) -> Arch:
        """
        Return the architecture of the binary package (main executable target).

        :return: architecture
        """
        return self._arch

    @property
    def platform(self) -> Platform:
        """
        Return the platform of the binary package (main exectuable target).

        :return: platform
        """
        return self._platform

    @staticmethod
    def auto(exe_file: Union[Path, str]) -> Optional['BinaryPackage']:
        """
        Take a file and try creating a BinaryPackage with it. The `exe_file` is
        the main executable file. From that the function will look for quokka,
        cmplog, dictionary files (in the same directory).

        :param exe_file: main target executable file
        :return: a binary package if `exe_file` if applicable
        """
        bin_f = Path(exe_file)

        # Exclude file if have one of the
        if bin_f.suffix in BinaryPackage.EXTENSION_BLACKLIST:
            return None

        # If do not exists
        if not bin_f.exists():
            return None

        # Make sure its an executable
        data = BinaryPackage._read_binary_infos(bin_f)
        if not data:
            return None

        bin_f.chmod(stat.S_IRWXU)  # make sure the binary is executable

        p = BinaryPackage(bin_f)
        p._platform, p._arch = data

        # Search for a Quokka file
        qfile1, qfile2 = Path(str(bin_f)+".Quokka"), Path(str(bin_f)+".quokka")
        if qfile1.exists():
            p._quokka = qfile1
        elif qfile2.exists():
            p._quokka = qfile2

        # Search for a graph file (containing callgraph)
        cfile = Path(str(bin_f)+".gt")
        if cfile.exists():
            p._callgraph = cfile

        # Search for a cmplog file if any
        cfile = Path(str(bin_f)+".cmplog")
        if cfile.exists():
            p._cmplog = cfile
            cfile.chmod(stat.S_IRWXU)  # make sure the cmplog binary is executable

        # Search for a dictionary file if any
        cfile = Path(str(bin_f)+".dict")
        if cfile.exists():
            p._dictionary = cfile

        return p

    @staticmethod
    def auto_directory(exe_file: Union[str, Path]) -> Optional['BinaryPackage']:
        """
        Create a BinaryPackage with all files it can find in the given
        directory. The difference with :py:meth:`BinaryPackage.auto` is
        that all additional files in the directory will be added to the
        package.

        :param exe_file: main executable in the directory
        :return: BinaryPackage if applicable
        """
        bin_f = Path(exe_file)

        p = BinaryPackage.auto(bin_f)

        if p is None:
            return None

        for file in bin_f.parent.iterdir():
            if file not in [p._main_bin, p._callgraph, p._quokka, p._cmplog, p._dictionary]:
                p.other_files.append(file)

    def make_package(self) -> Path:
        """
        Pack the BinaryPackage in a zip file.

        :return: Path to a .zip file containing the whole package
        """
        if self._package_file is not None:
            if self._package_file.exists():
                return self._package_file
        # Recreate a package
        fname = tempfile.mktemp(suffix=".zip")
        zip = zipfile.ZipFile(fname, "w")
        zip.write(self._main_bin, self._main_bin.name)
        if self._quokka:
            zip.write(self._quokka, self._quokka.name)
        if self._callgraph:
            zip.write(self._callgraph, self._callgraph.name)
        if self._cmplog:
            zip.write(self._cmplog, self._cmplog.name)
        if self._dictionary:
            zip.write(self._dictionary, self._dictionary.name)
        for file in self.other_files:
            zip.write(file)
        zip.close()
        return Path(fname)

    @staticmethod
    def _read_binary_infos(file: Path) -> Optional[Tuple[Platform, Arch]]:
        p = lief.parse(str(file))
        if not p:
            return None
        if not isinstance(p, lief.ELF.Binary):
            logging.warning(f"binary {file} not supported (only ELF at the moment)")
            return None

        # Determine the architecture of the binary
        mapping = {lief.ELF.ARCH.x86_64: Arch.X86_64,
                   lief.ELF.ARCH.i386: Arch.X86,
                   lief.ELF.ARCH.ARM: Arch.ARMV7,
                   lief.ELF.ARCH.AARCH64: Arch.AARCH64}
        arch = mapping.get(p.header.machine_type)

        # Determine the platform from its format
        mapping_elf = {lief.EXE_FORMATS.ELF: Platform.LINUX,
                       lief.EXE_FORMATS.PE: Platform.WINDOWS,
                       lief.EXE_FORMATS.MACHO: Platform.MACOS}
        # FIXME: differentiating between ELF (Linux, Android ..) and MACHO (MacOS, iOS..)
        fmt = mapping_elf.get(p.format)

        if arch and fmt:
            return fmt, arch
        else:
            return None

    @staticmethod
    def from_binary(name: str, binary: bytes, extract_dir: Path) -> 'BinaryPackage':
        """
        Convert the binary blob received as a BinaryPackage object. If its an archive,
        extract it and return the list of files. Files are extracted in /tmp. If
        directly an executable save it to a file and return its path. Also ensure
        the executable file is indeed executable in terms of permissions.

        :param name: name of executable, or executable name in archive
        :param binary: content
        :param extract_dir: Path: directory where files should be extracted
        :return: list of file paths

        :raise FileNotFoundError: if the mime type of the binary is not recognized
        """
        mime = magic.from_buffer(binary, mime=True)

        if mime in ['application/x-tar', 'application/zip']:
            map = {'application/x-tar': '.tar.gz', 'application/zip': '.zip'}
            tmp_file = Path(tempfile.mktemp(suffix=map[mime]))
            tmp_file.write_bytes(binary)          # write the archive in a file

            # Extract the archive in the right directory
            shutil.unpack_archive(tmp_file.as_posix(), extract_dir)  # unpack it in dst directory
            # Create the package object
            pkg = BinaryPackage.auto(Path(extract_dir) / name)
            if pkg is None:
                raise ValueError(f"Cannot create a BinaryPackage with {name}")
            for file in extract_dir.iterdir():
                if file not in [pkg.executable_path, pkg.callgraph, pkg.quokka, pkg.dictionary]:
                    pkg.other_files.append(file)
            return pkg
        elif mime in ['application/x-pie-executable', 'application/x-dosexec', 'application/x-mach-binary', 'application/x-executable', 'application/x-sharedlib']:
            program_path = extract_dir / name
            program_path.write_bytes(binary)
            program_path.chmod(stat.S_IRWXU)  # set the binary executable
            return BinaryPackage(program_path)
        else:
            raise FileNotFoundError(f"mimetype not recognized {mime}")



import json
from aenum import Enum, extend_enum
from pathlib import Path
from typing import Union
import base64

PathLike = Union[str, Path]
#: Union of a string or Path object

class State(Enum):
    """
    Running type of a fuzzing engine. It
    can be either running or idle.
    """
    RUNNING = 0
    IDLE = 1


class Platform(Enum):
    """
    Enum representing the platform.
    """
    ANY = 0
    LINUX = 1
    WINDOWS = 2
    MACOS = 3
    ANDROID = 4
    IOS = 5


class SeedType(Enum):
    """
    Type of an input. They can be plain input,
    crash input or hanging input.
    """
    INPUT = 0
    CRASH = 1
    HANG = 2


class ExecMode(Enum):
    """
    Execution mode for fuzzing engine. With ``AUTO``
    the fuzzer will automatically select, ``SINGLE_EXEC``
    is the normal fuzzing mode where the process stops
    at each iteration while ``PERSISTENT`` indicate the
    fuzzer to run in persistent mode.
    """
    AUTO = 0
    SINGLE_EXEC = 1
    PERSISTENT = 2


class FuzzMode(Enum):
    """
    Fuzzing mode, indicates the fuzzer whether the target
    is instrumented or not.
    """
    AUTO = 0
    INSTRUMENTED = 1
    BINARY_ONLY = 2


class CheckMode(Enum):
    """
    CheckMode is used to indicates a fuzzer how to run depdending on
    the context. ``CHECK_ALL`` is the normal bug, vulnerability discovery
    mode. Then ``ALERT_ONLY`` indicates the fuzzer to focus on SAST alerts.
    Then ``ALERT_ONE`` indicates the fuzzer to focus on a single alert which
    id should be provided through the configuration file.
    """
    CHECK_ALL = 0
    ALERT_ONLY = 1
    ALERT_ONE = 2

class CoverageMode(str, Enum):
    """
    Coverage metrics to use. Some fuzzing engines do support multiple coverage
    metrics, thus the enum indicates the one to use.
    """
    AUTO = "auto"
    BLOCK = "block"
    EDGE = "edge"
    PATH = "path"
    STATE = "state"

    @classmethod
    def _missing_(cls, val) -> 'CoverageMode':
        """ Method used to dynmically creating an entry """
        enum_name = val.upper().replace(" ", "_")
        if enum_name in cls.__members__:
            return cls.__members__[enum_name]
        return extend_enum(cls, enum_name, val)


class SeedInjectLoc(Enum):
    """
    Indicates the location where to inject inputs. It can either be
    on STDIN or ARGV.
    """
    STDIN = 0
    ARGV = 1


class Arch(Enum):
    """
    Architecture representation
    """
    X86 = 0
    X86_64 = 1
    ARMV7 = 2
    AARCH64 = 3


class LogLevel(Enum):
    """
    Enum representing the Log level, for fuzzers to send message logs
    to the broker.
    """
    DEBUG = 10
    INFO = 20
    WARNING = 2
    ERROR = 3
    CRITICAL = 4


class AlertData(object):
    """
    AlertData is data message that can be sent from fuzzing agents to
    the broker to indicates that an alert has been covered or validated.
    """
    def __init__(self, id: int, covered: bool, validated: bool, seed: bytes, address: int = 0):
        self.id: int = id
        #: Id of the alert
        self.covered: bool = covered
        #: True if the alert has been covered
        self.validated: bool = validated
        #: True if the alert has been validated
        self.seed: bytes = seed
        #: Input that reached or validated the alert
        self.address: int = address
        #: memory address of the alert


    @staticmethod
    def from_json(data: str) -> 'AlertData':
        """
        Convert an AlertData in json to an instance.

        :param data: json serialized alert
        :return: AlertData object
        """
        data = json.loads(data)

        return AlertData(data['id'], data['covered'], data['validated'], base64.b64decode(data['seed']), data['address'])

    def to_json(self) -> str:
        """
        Serialize the alert to JSON.

        :return: json serialized alert
        """
        return json.dumps({'id': self.id,
                           'covered': self.covered,
                           'validated': self.validated,
                           'seed': base64.b64encode(self.seed).decode(),
                           'address': self.address})


class FuzzingEngineInfo(object):
    """
    Class to represent a fuzzing engine metadata.
    It contains its name, version and the Python module
    where to load the descriptor and configuration object.
    """
    def __init__(self, name: str, version: str, pymodule: str):
        self.name: str = name
        #: Name of the engine
        self.version = version
        #: Version of the engine
        self.pymodule = pymodule
        #: Name of the python module

    @staticmethod
    def from_pb(pb) -> 'FuzzingEngineInfo':
        """
        Parse a protobuf object into a FuzzingEngineInfo object.

        :param pb: protobuf object
        :return: object
        """
        return FuzzingEngineInfo(pb.name, pb.version, pb.pymodule)



from typing import List

from pydantic import BaseModel

class InputEntry(BaseModel):
    engine: str
    number: int
    unique: int
    useless: int
    condition: int
    symread: int
    symwrite: int
    symjump: int

class CoverageEntry(BaseModel):
    engine: str
    number: int
    unique: int
    first: int
    total: int

class ExecEntry(BaseModel):
    engine: str
    dse: float
    smt: float
    replay: float
    total: float
    wait: float

class SeedSharingEntry(BaseModel):
    engine: str
    accepted: int
    rejected: int
    total: int
    ratio: float

class SmtEntry(BaseModel):
    engine: str
    sat: int
    unsat: int
    timeout: int
    total: int
    avg_query: float
    cov_sat_ratio: float
    branch_solved: int
    branch_not_solved: int


class CampaignStats(BaseModel):
    input_stats: List[InputEntry]
    coverage_stats: List[CoverageEntry]
    exec_stats: List[ExecEntry]
    seed_sharing_stats: List[SeedSharingEntry]
    smt_stats: List[SmtEntry]



# builtin imports
from typing import Callable
import time
import tempfile
import os
import logging
from pathlib import Path
from hashlib import md5

# third-party imports
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer


class Workspace(FileSystemEventHandler):

    AFLPP_WS_ENV_VAR = "AFLPP_WS"
    DEFAULT_WS_PATH = "aflpp_workspace"
    STATS_FILE = "fuzzer_stats"

    def __init__(self):
        self.observer = Observer()
        self.modif_callbacks = {}  # Map fullpath -> callback
        self.created_callbacks = {}
        self.root_dir = None
        self._setup_workspace()

    def _setup_workspace(self):
        ws = os.environ.get(self.AFLPP_WS_ENV_VAR, None)
        if ws is None:
            self.root_dir = (Path(tempfile.gettempdir()) / self.DEFAULT_WS_PATH) / str(time.time()).replace(".", "")
        else:
            self.root_dir = Path(ws)  # Use the one provided


        for d in [self.target_dir, self.input_dir, self.dynamic_input_dir, self.corpus_dir, self.crash_dir]:
            d.mkdir(parents=True)

        # Create dummy input file.
        # AFLPP requires that the initial seed directory is not empty.
        # TODO Is there a better approach to this?
        seed_path = self.input_dir / 'seed-dummy'
        seed_path.write_bytes(b'A')

    @property
    def target_dir(self):
        return self.root_dir / 'target'

    @property
    def input_dir(self):
        return self.root_dir / 'inputs' / 'initial'

    @property
    def dynamic_input_dir(self):
        return self.root_dir / 'inputs' / 'dynamic'

    @property
    def output_dir(self):
        return self.root_dir / 'outputs'

    @property
    def corpus_dir(self):
        return self.output_dir / 'main' / 'queue'

    @property
    def crash_dir(self):
        return self.output_dir / 'main' / 'crashes'

    @property
    def stats_dir(self):
        return self.output_dir / 'main'

    @property
    def stats_file(self):
        return self.stats_dir / self.STATS_FILE

    def on_modified(self, event):
        path = Path(event.src_path)
        if path.is_dir():
            return  # We don't care about directories
        if path.parent in self.modif_callbacks:
            self.modif_callbacks[path.parent](path)  # call the callback
        else:
            pass  # Do nothing at the moment

    def on_created(self, event):
        path = Path(event.src_path)
        if path.is_dir():
            return  # We don't care about directories
        if path.parent in self.created_callbacks:
            self.created_callbacks[path.parent](path)  # call the callback
        else:
            pass  # Do nothing at the moment

    def add_file_modification_hook(self, path: str, callback: Callable):
        self.observer.schedule(self, path=path, recursive=True)
        self.modif_callbacks[path] = callback

    def add_creation_hook(self, path: str, callback: Callable):
        self.observer.schedule(self, path=path, recursive=True)
        self.created_callbacks[path] = callback

    def start(self):
        self.observer.start()

    def stop(self):
        self.observer.stop()



# built-in imports
from abc import abstractmethod
from enum import Enum, auto
from pathlib import Path
from typing import Generator, Optional, Union, List, Dict, Tuple, Set
import os
import json
import sys
import subprocess
import logging
import re
from datetime import datetime, timedelta
from pydantic import BaseModel
import matplotlib.pyplot as plt

# third-party
from pastisbroker import BrokingMode
from pastisbroker.workspace import Workspace, WorkspaceStatus
from libpastis.types import SeedInjectLoc, SeedType
from tritondse.trace import QBDITrace
from tritondse import CoverageStrategy, GlobalCoverage, BranchSolvingStrategy, Config
from tritondse.coverage import CovItem

# local imports
from pastisbenchmark.replayer import ReplayType

class InputCovDelta(BaseModel):
    time_elapsed: float        # Time elapsed when input generated
    input_name: str            # Input name
    fuzzer: str                # The fuzzer that found that input

    # Trace generic info
    unique_items_covered_count: int            # The coverage of this one input (len(covered_items))
    unique_insts_covered_count: int            # Instruction coverage of that input

    # Local to the fuzzer
    fuzzer_new_items_covered: Set[CovItem]  # New items covered by THIS fuzzer
    fuzzer_coverage_sum: int            # Sum of covered items by THIS fuzzer at this point
    fuzzer_inst_coverage_sum: int       # Sum of unique instructions covered by THIS FUZZER at this point
    fuzzer_new_insts_covered: Set[int]  # Instr

    # Global accros all fuzzers of the campaign
    overall_new_items_covered: Set[CovItem]  # New items covered by this fuzzer agains ALL THE OTHERS
    overall_coverage_sum: int            # The total coverage of the fuzz campaign at this point
    overall_inst_coverage_sum: int       # Total instruction coverage at this point
    overall_new_insts_covered: Set[int]  # Instr


    def is_initial_input(self) -> bool:
        return self.fuzzer == "seeds"

    def is_triton_input(self) -> bool:
        return "TT" in self.fuzzer


class CampaignResult(object):

    SEED_FUZZER = "seeds"
    ALL_FUZZER = "all"

    QBDI_REPLAY_DIR = "replays_qbdi"
    LLVMPROFILE_REPLAY_DIR = "replays_llvmprof"
    REPLAYS_DELTA = "replays_delta"
    CLIENT_STATS = "clients-stats.json"
    COVERAGE_DIR = "coverages"

    def __init__(self, workspace: Union[Path, str]):
        self.workspace = Workspace(Path(workspace))
        # Stat items
        self.fuzzers_items = {}     # fuzzer_name -> List[StatItem]
        self.fuzzers_coverage = {}  # fuzzer_name -> Coverage
        self.fuzzers_config = {}    # fuzzer_name -> Union[str, Config]
        self._load_fuzzer_configs()
        self._all_items = []

        # initialize directories
        self._init_directories()

        # Load fuzzers configuration

        self.mode = self.load_broking_mode(self.workspace)

    def _load_fuzzer_configs(self):
        f = self.workspace.root / self.CLIENT_STATS
        data = json.loads(f.read_text())
        for client in data:
            id = client['strid']
            conf = client['engine_args']
            if self.is_triton(id):
                self.fuzzers_config[id] = Config.from_json(conf)
            else:
                self.fuzzers_config[id] = conf

    def has_honggfuzz(self):
        for fuzz in self.fuzzers_items.keys():
            if "HF" in fuzz:
                return True
        return False

    @staticmethod
    def is_triton(fuzzer: str) -> bool:
        return "TT" in fuzzer

    @property
    def is_full_duplex(self) -> bool:
        return bool(self.mode == BrokingMode.FULL)

    @property
    def is_half_duplex(self) -> bool:
        return bool(self.mode == BrokingMode.NO_TRANSMIT)

    @property
    def slug_name(self) -> str:
        data = ["AFL" if any(("AFLPP" in x for x in self.fuzzers_config)) else "",
                "HF" if any(("HF" in x for x in self.fuzzers_config)) else "",
                "TT" if any(("TT" in x for x in self.fuzzers_config)) else ""]
        if self.is_full_duplex:
            return f"PASTIS[{'|'.join(x for x in data if x)}]"
        else:
            return f"U[{'|'.join(x for x in data if x)}]"

    @property
    def results(self):
        return self.fuzzers_items.items()

    @property
    def delta_items(self) -> Generator[InputCovDelta, None, None]:
        yield from self._all_items

    def _init_directories(self):
        if not self.replay_delta_dir.exists():
            self.replay_delta_dir.mkdir()

    def _init_fuzzer_coverage(self, fuzzer_name: str) -> None:
        if fuzzer_name in self.fuzzers_coverage:
            return None

        # consider seed inputs as common to all
        if self.SEED_FUZZER in self.fuzzers_coverage:
            cov = self.fuzzers_coverage[self.SEED_FUZZER].clone()
        else: # else create an empty coverage file
            cov = GlobalCoverage(CoverageStrategy.EDGE, BranchSolvingStrategy.ALL_NOT_COVERED)

        self.fuzzers_coverage[fuzzer_name] = cov

    @staticmethod
    def load_broking_mode(workspace: Workspace) -> BrokingMode:
        data = json.loads(workspace.config_file.read_text())
        return BrokingMode[data['broker_mode']]

    @property
    def replay_delta_dir(self) -> Path:
        return self.workspace.root / self.REPLAYS_DELTA

    def has_delta_files(self) -> bool:
        return len(list(self.replay_delta_dir.iterdir())) != 0

    def replay_ok(self) -> bool:
        return len(list(self.replay_delta_dir.iterdir())) != 0

    @staticmethod
    def parse_filename(filename) -> Optional[tuple]:
        ref = datetime.strptime("0:00:00.00", "%H:%M:%S.%f")
        if re.match("^\d{4}-\d{2}-\d{2}", filename):  # start by the date
            date, time, elapsed, fuzzer_id, hash = filename.split("_")
            date = datetime.strptime(f"{date}_{time}", "%Y-%m-%d_%H:%M:%S")
            spl = elapsed.split(":")
            if len(spl) == 4:
                elapsed = int(spl[0][:-1])*(3600*24)
                elapsed += (datetime.strptime(":".join(spl[1:]), "%H:%M:%S.%f") - ref).total_seconds()
            else:
                elapsed = (datetime.strptime(elapsed, "%H:%M:%S.%f") - ref).total_seconds()
            hash = hash.split(".")[0]
            return date, elapsed, fuzzer_id, hash
        else:
            return None

    def _iter_sorted(self, path: Path):
        files = {None: []}
        for file in path.iterdir():
            res = self.parse_filename(file.name)
            if res is None:
                files[None].append(file)
            else:
                date, elapsed, fuzzer, hash = res
                if elapsed in files:
                    logging.warning(f"two files with same elapsed time: {files[elapsed]} | {file.name}")
                files[elapsed] = file

        # First yield initial seeds
        init_seeds = files.pop(None)
        yield from init_seeds

        # Then iterate file sorted by elapsed time
        for k in sorted(files):
            yield files[k]

    def load(self, type: ReplayType = ReplayType.qbdi) -> None:
        logging.info(f"load in {type.name} [mode:{self.mode.name}]")
        if self.has_delta_files():
            items = self.load_delta_directory()
            self.load_delta(items)
            self.load_coverage()  # If delta, coverage files shall also be present
        elif type == ReplayType.qbdi:
            items = self.load_qbdi()
            self.load_delta(items)
            self.save_coverage()  # Once loaded save coverage files
        elif type == ReplayType.llvm_profile:
            self.load_llvmprofile()
        else:
            assert False

    def load_qbdi(self) -> List[Tuple[str, InputCovDelta]]:
        logging.info("load qbdi trace files")
        folder = self.workspace.root / self.QBDI_REPLAY_DIR
        total = sum(1 for _ in self._iter_sorted(folder))

        items = []

        # initialize a "all" fuzzer in all cases
        self._init_fuzzer_coverage(self.ALL_FUZZER)

        for i, file in enumerate(self._iter_sorted(folder)):
            # parse name
            print(f"[{i+1}/{total}] {file}\r", file=sys.stderr, end="")
            meta = self.parse_filename(file.name)

            # Get the fuzzer name (and coverage)
            if meta is None:
                fuzzer = self.SEED_FUZZER
            else:
                fuzzer = meta[2]

            self._init_fuzzer_coverage(fuzzer)

            cov = QBDITrace.from_file(file).coverage

            # Compute differences to know what has been covered
            if self.is_half_duplex:
                fuzzer_coverage = self.fuzzers_coverage[fuzzer]   # get the fuzzer coverage at this point
                local_new_items = cov - fuzzer_coverage
                local_new_instrs = cov.covered_instructions.keys() - fuzzer_coverage.covered_instructions.keys()
                fuzzer_coverage.merge(cov)   # Merge the new coverage
                cov_sum = fuzzer_coverage.unique_covitem_covered
                inst_sum = fuzzer_coverage.unique_instruction_covered
            else:  # FULL DUPLEX
                local_new_items, local_new_instrs, cov_sum, inst_sum = set(), set(), 0, 0


            # Update coverage and info OVERALL
            all_coverage = self.fuzzers_coverage[self.ALL_FUZZER]  # get the overall coverage
            overall_new_items = cov - all_coverage                 # compute new items
            overall_new_instrs = cov.covered_instructions.keys() - all_coverage.covered_instructions.keys()
            all_coverage.merge(cov)                                # Merge all of them (all fuzzers all together)


            # Create an InputCovDelta
            statitem = InputCovDelta(
                time_elapsed=0 if meta is None else meta[1],
                input_name=file.name,
                fuzzer=fuzzer,
                # Trace generic info
                unique_items_covered_count=cov.unique_covitem_covered,         # The coverage of this one input (len(covered_items))
                unique_insts_covered_count=cov.unique_instruction_covered, # Instruction coverage of that input
                # Local to the fuzzer
                fuzzer_new_items_covered=local_new_items, # New items covered by THIS fuzzer
                fuzzer_coverage_sum=cov_sum,              # Sum of covered items by THIS fuzzer at this point
                fuzzer_inst_coverage_sum=inst_sum,        # Sum of unique instructions covered by THIS FUZZER at this point
                fuzzer_new_insts_covered=local_new_instrs,# Instr
                # Global accros all fuzzers of the campaign
                overall_new_items_covered=overall_new_items,              # New items covered by this fuzzer agains ALL THE OTHERS
                overall_coverage_sum=all_coverage.unique_covitem_covered, # The total coverage of the fuzz campaign at this point
                overall_inst_coverage_sum=all_coverage.unique_instruction_covered, # Total instruction coverage at this point
                overall_new_insts_covered=overall_new_instrs
            )

            items.append((fuzzer, statitem))

            # Write the delta file
            with open(self.replay_delta_dir / (file.name+".json"), 'w') as f:
                f.write(statitem.json())
        return items

    def load_llvmprofile(self) -> None:
        # TODO: to implement
        pass

    def load_delta_directory(self) -> List[Tuple[str, InputCovDelta]]:
        logging.info("load delta directory")

        items = []

        for file in self._iter_sorted(self.replay_delta_dir):
            meta = self.parse_filename(file.name)

            # Get the fuzzer name (and coverage)
            if meta is None:
                fuzzer = self.SEED_FUZZER
            else:
                fuzzer = meta[2]

            delta = InputCovDelta.parse_file(file)
            items.append((fuzzer, delta))
        return items

    def load_delta(self, items: List[Tuple[str, InputCovDelta]]) -> None:
        self._all_items = [x[1] for x in items]
        self.fuzzers_items[self.SEED_FUZZER] = []
        self.fuzzers_items[self.ALL_FUZZER] = []

        for fuzzer, item in items:
            if fuzzer not in self.fuzzers_items:
                self.fuzzers_items[fuzzer] = []

            self.fuzzers_items[fuzzer].append(item)

            if fuzzer != self.SEED_FUZZER:  # Add in ALL, all that are not seed ones
                self.fuzzers_items[self.ALL_FUZZER].append(item)

    def save_coverage(self):
        cov_dir = self.workspace.root / self.COVERAGE_DIR
        if not cov_dir.exists():
            cov_dir.mkdir()

        for fuzzer_name, cov in self.fuzzers_coverage.items():
            covfile = cov_dir / (fuzzer_name + ".ttgcov")
            cov.to_file(covfile)

            if cov.covered_instructions:  # if we have instructions covered save file in Lightouse format
                covfile = cov_dir / (fuzzer_name + ".cov")
                with open(covfile, "w") as f:
                    f.write("\n".join(f"{x:#08x}" for x in cov.covered_instructions.keys()))
                    f.write("\n")

    def load_coverage(self):
        cov_dir = self.workspace.root / self.COVERAGE_DIR

        for fuzzer_name in self.fuzzers_items.keys():
            covfile = cov_dir / (fuzzer_name + ".ttgcov")
            if not covfile.exists():
                logging.error(f"can't find coverage of {fuzzer_name}")
            else:
                self.fuzzers_coverage[fuzzer_name] = GlobalCoverage.from_file(covfile)

    def print_stats(self):
        overall_coverage = self.fuzzers_coverage[self.ALL_FUZZER]
        for fuzzer, results in self.fuzzers_items.items():

            tot_inps = sum(len(x) for x in self.fuzzers_items.values())
            tot_edge = overall_coverage.unique_covitem_covered

            print("-----------------------")
            print(f"Fuzzer: {fuzzer}")
            cov = self.fuzzers_coverage[fuzzer]
            print(f"inputs: {len(results)}, {len(results)/tot_inps:.0%}")
            print(f"edges: {cov.unique_covitem_covered} {cov.unique_covitem_covered/tot_edge:.0%)}")
            print(f"instructions: {cov.unique_instruction_covered}")
            print("-----------------------")
            print(f"Total inputs: {tot_inps}")
            print(f"Total edges: {tot_edge}")



import os
from pathlib import Path
from typing import Optional
import subprocess

from .driver import HonggfuzzDriver
from .replay import Replay
from .honggfuzz import HonggfuzzProcess, HonggfuzzNotFound
from .workspace import Workspace

__version__ = "1.0.0"

# Honggfuzz env variables
HFUZZ_ENV_VAR = "HFUZZ_WS"
HFUZZ_PATH_VAR = "HFUZZ_PATH"
HFUZZ_THREADS = "HFUZZ_THREADS"


def spawn_online_honggfuzz(workspace: Optional[Path], hf_path: Optional[str], port: int = 5555, threads: int=0):
    env = os.environ
    env[HFUZZ_ENV_VAR] = str(workspace.absolute())
    if hf_path:
        env[HFUZZ_PATH_VAR] = str(hf_path)
    if threads:
        env[HFUZZ_THREADS] = str(threads)
    cmd_line_honggfuzz = ["pastis-honggfuzz", "online", "-p", f"{port}"]
    subprocess.Popen(cmd_line_honggfuzz, env=env, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)




# builtin imports
import logging
import os
import re
import signal
import subprocess
import time
from typing import Optional, Union
from pathlib import Path

# third-party imports
import shutil
from libpastis.types import ExecMode, FuzzMode

# Local imports
from .workspace import Workspace


class AFLPPNotFound(Exception):
    """ Issue raised on """
    pass


class AFLPPProcess:

    AFLPP_ENV_VAR = "AFLPP_PATH"
    BINARY = "afl-fuzz"
    STAT_FILE = "fuzzer_stats"
    VERSION = "master"

    def __init__(self, path: str = None):

        self.__path = self.find_alfpp_binary(path)
        if self.__path is None:
            raise FileNotFoundError("Can't find AFL++ path (afl-fuzz)")

        self.__process = None
        self.__logfile = None

    @staticmethod
    def find_alfpp_binary(root_dir: Union[Path, str]) -> Optional[Path]:
        if root_dir:
            bin_path = Path(root_dir) / AFLPPProcess.BINARY
            return bin_path if bin_path.exists() else None
        else:
            aflpp_path = os.environ.get(AFLPPProcess.AFLPP_ENV_VAR)
            return Path(aflpp_path) / 'afl-fuzz' if aflpp_path else shutil.which(AFLPPProcess.BINARY)

    def start(self, target: str, target_arguments: list[str], workspace: Workspace, exmode: ExecMode, fuzzmode: FuzzMode, stdin: bool, engine_args: str, cmplog: Optional[str] = None, dictionary: Optional[str] = None):
        # Check that we have '@@' if input provided via argv
        if not stdin:
            if "@@" not in target_arguments:
                logging.error(f"seed provided via ARGV but can't find '@@' on program argv")
                return
        # Build target command line.
        target_cmdline = f"{target} {' '.join(target_arguments)}"

        # Build fuzzer arguments.
        # NOTE: Assuming the target receives inputs from stdin.
        aflpp_arguments = ' '.join([
            re.sub(r"\s", " ", engine_args),  # Any arguments coming right from the broker (remove \r\n)
            f"-Q" if fuzzmode == FuzzMode.BINARY_ONLY else "",
            f"-M main", # Master MODE, seed distribution is ensured by the broker
            f"-i {workspace.input_dir}",
            f"-F {workspace.dynamic_input_dir}",
            f"-o {workspace.output_dir}",
            f"-c {cmplog}" if cmplog is not None else "",
            f"-x {dictionary}" if dictionary is not None else ""
        ])

        # Export environmental variables.
        os.environ["AFL_NO_UI"] = "1"
        os.environ["AFL_QUIET"] = "1"
        os.environ["AFL_IMPORT_FIRST"] = "1"
        os.environ["AFL_AUTORESUME"] = "1"

        # NOTE This prevents having to configure the system before running
        #      AFL++.
        # TODO Should we skip these steps?
        os.environ["AFL_SKIP_CPUFREQ"] = "1"
        os.environ["AFL_I_DONT_CARE_ABOUT_MISSING_CRASHES"] = "1"

        # Build fuzzer command line.
        aflpp_cmdline = f'{self.__path} {aflpp_arguments} -- {target_cmdline}'

        logging.info(f"Run AFL++: {aflpp_cmdline}")
        logging.debug(f"\tWorkspace: {workspace.root_dir}")

        # Remove empty strings when converting the command to a list.
        command = list(filter(None, aflpp_cmdline.split(' ')))

        # Open logfile (stdout will be redirected to this file).
        self.__logfile = open(workspace.root_dir / 'logfile.log', 'w')

        # Create a new fuzzer process and set it apart into a new process group.
        self.__process = subprocess.Popen(command, cwd=str(workspace.root_dir), preexec_fn=os.setsid, stdout=self.__logfile)

        logging.debug(f'Process pid: {self.__process.pid}')

    @property
    def instanciated(self):
        return self.__process is not None

    def stop(self):
        if self.__process:
            logging.debug(f'Stopping process with pid: {self.__process.pid}')
            os.killpg(os.getpgid(self.__process.pid), signal.SIGTERM)
        else:
            logging.debug(f"AFL++ process seems already killed")

        if self.__logfile:
            self.__logfile.close()

    def wait(self):
        while not self.instanciated:
            time.sleep(0.1)
        self.__process.wait()
        logging.info(f"Fuzzer terminated with code : {self.__process.returncode}")

    @staticmethod
    def aflpp_environ_check() -> bool:
        return os.environ.get(AFLPPProcess.AFLPP_ENV_VAR) is not None



# Built-in imports
from typing import Tuple, List, Dict
from pathlib import Path
import logging
import time
import inspect

# Third-party imports
from libpastis.types import FuzzingEngineInfo, Arch, LogLevel, ExecMode, CheckMode, CoverageMode, SeedType, Platform
from libpastis import FuzzingEngineDescriptor


class PastisClient(object):
    """
    Utility class holding all information related to
    a client connected to the broker.
    """

    def __init__(self, id: int, netid: bytes, engines: List[FuzzingEngineInfo], arch: Arch, cpus: int, memory: int, hostname: str, platform: Platform):
        # All this attributes are assigned once and for all
        self.id = id
        self.netid = netid
        self.engines = engines
        self.arch = arch
        self.cpus = cpus
        self.memory = memory
        self.hostname = hostname
        self.platform = platform

        self.logger = None

        # Runtime properties (reset at avery send_start)
        self._program = None
        self._running = False
        self._engine = None  # FuzzingEngineDescriptor
        self._engine_args = None
        self._coverage_mode = None
        self._exec_mode = None
        self._check_mode = None
        self._seeds_received = set()  # Seed sent to the client
        self._seeds_submitted = set()  # Seed submitted by the client
        self.target = None  # target in case of slicing
        self.target_validated = False

        # Runtime telemetry stats
        self.exec_per_sec = None
        self.total_exec = None
        self.cycle = None
        self.timeout = None
        self.coverage_block = None
        self.coverage_edge = None
        self.coverage_path = None
        self.last_cov_update = None

        # seed stats
        self.input_submitted_count = 0
        self.crash_submitted_count = 0
        self.timeout_submitted_count = 0
        self.seed_first = 0

        # SAST parameters
        self.alert_covered = set()
        self.alert_covered_first = 0
        self.alert_validated = set()
        self.alert_validated_first = 0

        # time series
        self._timeline_seeds = []  # List[Tuple[float, int, typ]]  # history of submission
        # self._timeline_coverage = []

    def configure_logger(self, log_dir, colorid: int):
        if self.logger is None:
            self.logger = logging.getLogger(f"\033[7m\033[{colorid}m[{self.strid}]\033[0m")

            # Add a file handler
            hldr = logging.FileHandler(log_dir/f"{self.strid}.log")
            hldr.setLevel(logging.DEBUG)
            hldr.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s]: %(message)s"))
            self.logger.addHandler(hldr)

    @property
    def strid(self):
        return f"CLI-{self.id}-{self._engine.SHORT_NAME if self._engine else 'N/A'}"

    def is_new_seed(self, seed: bytes) -> bool:
        """
        Return true if the seed has never been sent to a client

        :param seed: seed bytes
        :return: True if never sent to client
        """
        return seed not in self._seeds_received and seed not in self._seeds_submitted

    def add_peer_seed(self, seed: bytes) -> None:
        self._seeds_received.add(seed)

    def add_own_seed(self, seed: bytes) -> None:
        self._seeds_submitted.add(seed)

    def is_running(self) -> bool:
        return self._running

    def is_idle(self) -> bool:
        return not self._running

    def log(self, level: LogLevel, message: str):
        # Get the function in the logger (warning, debug, info) and call it with message
        if self.logger is None:  # Client has not yet been configured
            getattr(logging, level.name.lower())(message)
        else:  # Log in client logger
            getattr(self.logger, level.name.lower())(message)

    @property
    def package_name(self) -> str:
        return self._program

    @property
    def engine(self):
        return self._engine

    @property
    def coverage_mode(self):
        return self._coverage_mode

    @property
    def exec_mode(self):
        return self._exec_mode

    @property
    def check_mode(self):
        return self._check_mode

    def set_stopped(self):
        """ Flush runtime data (keep stats) """
        self._running = False
        self._engine = None
        self._coverage_mode = None
        self._exec_mode = None
        self._check_mode = None

    def set_running(self, program: str, engine: FuzzingEngineDescriptor, covmode: CoverageMode, exmode: ExecMode,
                    ckmode: CheckMode, engine_args: str = None):
        self._program = program
        self._running = True
        self._engine = engine
        self._coverage_mode = covmode
        self._exec_mode = exmode
        self._check_mode = ckmode
        self._engine_args = engine_args
        self._seeds_received = set()  # Seed sent to the client
        self._seeds_submitted = set()  # Seed submitted by the client

    def is_supported_engine(self, engine: FuzzingEngineDescriptor) -> bool:
        for e in self.engines:
            if e.name == engine.NAME:
                return True
        return False

    def to_dict(self) -> Dict:
        return {
            "id": self.id,
            "strid": self.strid,
            "engines": [x.name for x in self.engines],
            "arch": self.arch.name,
            "cpus": self.cpus,
            "memory": self.memory,
            "hostname": self.hostname,
            "platform": self.platform.name,
            "engine": self._engine.NAME if self._engine else "",
            "engine_args": self._engine_args,
            "coverage_mode": self._coverage_mode.name,
            "exec_mode": self._exec_mode.name,
            "check_mode": self._check_mode.name,
            "seed_received_count": len(self._seeds_received),
            "exec_per_sec": self.exec_per_sec,
            "total_exec": self.total_exec,
            "cycle": self.cycle,
            "timeout": self.timeout,
            "coverage_block": self.coverage_block,
            "coverage_edge": self.coverage_edge,
            "coverage_path": self.coverage_path,
            "last_cov_update": self.last_cov_update,
            "input_submitted_count": self.input_submitted_count,
            "crash_submitted_count": self.crash_submitted_count,
            "timeout_submitted_count": self.timeout_submitted_count,
            "seed_first": self.seed_first,
            "alert_covered": list(self.alert_covered),
            "alert_covered_first": self.alert_covered_first,
            "alert_validated": list(self.alert_validated),
            "alert_validated_first": self.alert_validated_first,
        }

    def add_covered_alert(self, a_id: int, cov: bool, cov_first: bool, val: bool, val_first: bool):
        if cov:
            self.alert_covered.add(a_id)
        if cov_first:
            self.alert_covered_first += 1
        if val:
            self.alert_validated.add(a_id)
        if val_first:
            self.alert_validated_first += 1



# built-in imports
import json
import csv
from pathlib import Path
from typing import List, Dict, Tuple, Union


class SASTAlert:
    """
    Class representing an alert in a somewhat abstract SAST tool. Its
    used to perform alert driven testing.
    """
    def __init__(self):
        # Static alert data
        self.id: int = -1
        #: Unique ID of the alert
        self.type: str = ""
        #: Type of the alert BoF, UaF (in the convention of the SAST)
        self.params: list = []
        #: Additional parameters of the alert (list)
        self.taxonomy: str = ""
        #: Taxonomy of the alert (e.g: CWE, CVE, MISRA checker, ..)
        self.severity: str = ""
        #: Severity of the alert (e.g: Review, Error, Critical ..)
        self.file: str = ""
        #: Source file impacted
        self.line: int = -1
        #: line of code (in the file)
        self.function: str = ""
        #: Function impacted
        self.raw_line: str = ""
        #: Raw alert extract taken from the report (in its own format)

        # Analysis results
        self.covered = False
        #: Coverage: True if the alert has been covered (path leading there)
        self.validated = False
        #: Validation: True if the alert has been validated (as a true positive by a checker)
        self.uncoverable = False
        #: Reachability: True if the alert cannot be reached by any paths

    @staticmethod
    def from_json(data: dict) -> 'SASTAlert':
        """
        Create a SASTAlert object from the JSON data provided.

        :param data: JSON data of the alert
        :return: SASTAlert instance, initialized with the JSON
        """
        alert = SASTAlert()
        for name in ["id", "type", "params", "taxonomy", "severity", "file", "line", "function", "raw_line",
                     "covered", "validated", "uncoverable"]:
            val = data.get(name)
            if val:
                setattr(alert, name, val)
        return alert


    def to_dict(self) -> dict:
        """
        Export the alert attribute to a valid JSON dictionnary
        that can be written to file.

        :return: JSON dict of the alert serialized
        """
        return {x: getattr(self, x) for x in ["id", "type", "params", "taxonomy", "severity", "file", "line", "function",
                                              "raw_line", "covered", "validated", "uncoverable"]}

    def __repr__(self):
        return f"<Alert id:{self.id}: {self.type} {self.function}:{self.line} ({Path(self.file).name})>"



class SASTReport:
    """
    SAST report. Manages a list of SAST alerts taken from a report.
    """

    def __init__(self):
        self.alerts: Dict[int, SASTAlert] = {}
        #: Dictionnary of alerts indexed by their ID


    def iter_alerts(self) -> List[SASTAlert]:
        """
        Iterate all the alerts of the report.
        :return: list of alerts
        """
        return list(self.alerts.values())


    def all_alerts_validated(self) -> bool:
        """
        Checks if all alerts have been validated (and thus covered)

        :return: True if all alerts are covered and vulns validated
        """
        for alert in self.iter_alerts():
            if not alert.covered:
                return False
        return True


    def add_alert(self, alert: SASTAlert) -> None:
        """
        Add an alert in the report. This function is solely
        meant to be used by the report parser

        :param alert: Alert object to add in the report
        """
        self.alerts[alert.id] = alert


    @staticmethod
    def from_file(file: Union[str, Path]) -> 'SASTReport':
        """
        Parse the given file into a SAST report object.

        :param file: path to report
        :return: SASTReport object
        """
        data = Path(file).read_bytes()
        return SASTReport.from_json(data)


    @staticmethod
    def from_json(data: Union[str, bytes]) -> 'SASTReport':
        """
        Parse the given string into a SAST report object.

        :param data: serialized report in JSON
        :return: SASTReport object
        """
        data = json.loads(data)
        report = SASTReport()
        for it in data:
            a = SASTAlert.from_json(it)
            report.add_alert(a)
        return report


    def to_json(self) -> str:
        """
        Export the current state of the alerts within a JSON dictionnary.

        :return: JSON serialized report
        """
        return json.dumps([x.to_dict() for x in self.alerts.values()], indent=2)


    def write(self, out_file) -> None:
        """
        Export the current state of the alerts within a JSON dictionary.

        :param out_file: Output file path
        """
        with open(out_file, "w") as f:
            f.write(self.to_json())


    def get_stats(self) -> Tuple[int, int, int]:
        """
        Get stats about the report. The results is a triple
        with the number of alerts covered, validated and total.

        :return: triple of covered, validated, totoal number of alerts
        """
        covered = 0
        validated = 0
        total = 0
        for alert in self.alerts.values():
            covered += int(alert.covered)
            validated += int(alert.validated)
            total += 1
        return covered, validated, total

    def write_csv(self, file: Path) -> None:
        """
        Write the report as a csv into the given file.

        :param file: CSV file to write
        """
        with open(file, 'w', newline='') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=['id', 'type', 'covered', 'validated'])
            writer.writeheader()
            for a in self.iter_alerts():
                writer.writerow({'id': a.id,
                                 'type': a.type,
                                 'covered': a.covered,
                                 'validated': a.validated})



import json
from pathlib import Path
from typing import Iterator, Generator
import shutil
import stat
from enum import Enum, auto

from libpastis.types import SeedType, PathLike
from libpastis import SASTReport


class WorkspaceStatus(Enum):
    NOT_STARTED = auto()
    RUNNING = auto()
    FINISHED = auto()


class Workspace(object):
    INPUT_DIR = "corpus"
    HANGS_DIR = "hangs"
    CRASH_DIR = "crashes"
    LOG_DIR = "logs"
    BINS_DIR = "binaries"
    ALERTS_DIR = "alerts_data"
    SEED_DIR = "seeds"

    SAST_REPORT_COPY = "sast-report.bin"
    CSV_FILE = "results.csv"
    TELEMETRY_FILE = "telemetry.csv"
    CLIENTS_STATS = "clients-stats.json"
    LOG_FILE = "broker.log"
    STATUS_FILE = "STATUS"
    RUNTIME_CONFIG_FILE = "config.json"
    COVERAGE_HISTORY = "coverage-history.csv"

    def __init__(self, directory: Path, erase: bool = False):
        self.root = directory

        if erase:  # If want to erase the whole workspace
            shutil.rmtree(self.root)

        # Create the base directory structure
        if not self.root.exists():
            self.root.mkdir()
        for s in [self.INPUT_DIR, self.CRASH_DIR, self.LOG_DIR, self.HANGS_DIR, self.BINS_DIR, self.SEED_DIR]:
            p = self.root / s
            if not p.exists():
                p.mkdir()

        # If no status file is found create one
        status_file = Path(self.root / self.STATUS_FILE)
        self._status = WorkspaceStatus.NOT_STARTED
        if not status_file.exists():
            status_file.write_text(self._status.name)
        else:
            self._status = WorkspaceStatus[status_file.read_text()]

    def initialize_runtime(self, binaries_dir: PathLike, params: dict):
        # First copy binary files in workspace if different directories
        if self.root / self.BINS_DIR != binaries_dir:
            shutil.copytree(binaries_dir, self.root / self.BINS_DIR, dirs_exist_ok=True)
        # Save runtime configuration
        config = self.root / self.RUNTIME_CONFIG_FILE
        config.write_text(json.dumps(params))

    def iter_corpus_directory(self, typ: SeedType) -> Generator[Path, None, None]:
        dir_map = {SeedType.INPUT: self.INPUT_DIR, SeedType.CRASH: self.CRASH_DIR, SeedType.HANG: self.HANGS_DIR}
        dir = self.root / dir_map[typ]
        for file in dir.iterdir():
            yield file

    def iter_initial_corpus_directory(self) -> Generator[Path, None, None]:
        for file in (self.root / self.SEED_DIR).iterdir():
            yield file

    def count_corpus_directory(self, typ: SeedType) -> int:
        return sum(1 for _ in self.iter_corpus_directory(typ))

    @property
    def status(self) -> WorkspaceStatus:
        return self._status

    @status.setter
    def status(self, value: WorkspaceStatus) -> None:
        self._status = value
        Path(self.root / self.STATUS_FILE).write_text(value.name)

    @property
    def telemetry_file(self) -> Path:
        return self.root / self.TELEMETRY_FILE

    @property
    def clients_stat_file(self) -> Path:
        return self.root / self.CLIENTS_STATS

    @property
    def sast_report_file(self) -> Path:
        return self.root / self.SAST_REPORT_COPY

    @property
    def csv_result_file(self) -> Path:
        return self.root / self.CSV_FILE

    @property
    def log_directory(self) -> Path:
        return self.root / self.LOG_DIR

    @property
    def broker_log_file(self) -> Path:
        return self.root / self.LOG_FILE

    @property
    def config_file(self) -> Path:
        return self.root / self.RUNTIME_CONFIG_FILE

    @property
    def coverage_history(self) -> Path:
        return self.root / self.COVERAGE_HISTORY

    def add_binary(self, binary_path: Path) -> Path:
        """
        Add a binary in the workspace directory structure.

        :param binary_path: Path of the executable to copy
        :return: the final executable file path
        """
        dst_file = self.root / self.BINS_DIR / binary_path.name
        if dst_file.absolute() != binary_path.absolute():  # If not already in the workspace copy them in workspace
            dst_file.write_bytes(binary_path.read_bytes())
            dst_file.chmod(stat.S_IRWXU)  # Change target mode to execute.
        return dst_file

    def add_binary_data(self, name: str, content: bytes) -> Path:
        """
        Add a binary in the workspace directory structure.

        :param name: Name of the executable file
        :param content: Content of the executable
        :return: the final executable file path
        """
        dst_file = self.root / self.BINS_DIR / name
        dst_file.write_bytes(content)
        dst_file.chmod(stat.S_IRWXU)  # Change target mode to execute.
        return dst_file

    def add_sast_report(self, report: SASTReport) -> Path:
        f = self.root / self.SAST_REPORT_COPY
        report.write(f)
        return f

    @property
    def binaries(self) -> Generator[Path, None, None]:
        for file in (self.root / self.BINS_DIR).iterdir():
            yield file

    def initialize_alert_corpus(self, report: SASTReport) -> None:
        """ Create a directory for each alert where to store coverage / vuln corpus """
        p = self.root / self.ALERTS_DIR
        p.mkdir(exist_ok=True)
        for alert in report.iter_alerts():
            a_dir = p / str(alert.id)
            a_dir.mkdir(exist_ok=True)

    def save_alert_seed(self, id: int, name: str, data: bytes) -> None:
        p = ((self.root / self.ALERTS_DIR) / str(id)) / name
        p.write_bytes(data)

    def save_seed_file(self, typ: SeedType, file: Path, initial: bool = False) -> None:
        dir_map = {SeedType.INPUT: self.INPUT_DIR, SeedType.CRASH: self.CRASH_DIR, SeedType.HANG: self.HANGS_DIR}
        if initial:
            out = self.root / self.SEED_DIR / file.name
        else:
            out = self.root / dir_map[typ] / file.name
        if str(file) != str(out):
            shutil.copy(str(file), str(out))

    def save_seed(self, typ: SeedType, name: str, data: bytes) -> None:
        dir_map = {SeedType.INPUT: self.INPUT_DIR, SeedType.CRASH: self.CRASH_DIR, SeedType.HANG: self.HANGS_DIR}
        out = self.root / dir_map[typ] / name
        out.write_bytes(data)



# built-in imports
from typing import List, Tuple
import os
import time
import logging
from pathlib import Path
import threading
import platform
import json
import queue

# third-party imports
from triton               import MemoryAccess, CPUSIZE

# Pastis & triton imports
import pastisdse
from tritondse            import Config, Program, CleLoader, CoverageStrategy, SymbolicExplorator, \
                                 SymbolicExecutor, ProcessState, ExplorationStatus, SeedStatus, ProbeInterface, \
                                 Workspace, Seed, CompositeData, SeedFormat, QuokkaProgram
from tritondse.sanitizers import FormatStringSanitizer, NullDerefSanitizer, UAFSanitizer, IntegerOverflowSanitizer, mk_new_crashing_seed
from tritondse.types      import Addr, Edge, SymExType, Architecture, Platform
from libpastis import ClientAgent, BinaryPackage, SASTReport
from libpastis.types      import SeedType, FuzzingEngineInfo, ExecMode, CoverageMode, SeedInjectLoc, CheckMode, LogLevel, AlertData, FuzzMode
from tritondse.trace      import QBDITrace, TraceException
from tritondse.worklist import FreshSeedPrioritizerWorklist, WorklistAddressToSet


class PastisDSE(object):

    INPUT_FILE_NAME = "input_file"
    STAT_FILE = "pastidse-stats.json"

    RAMDISK = "/mnt/ramdisk"
    TMP_SEED = "seed.seed"
    TMP_TRACE = "result.trace"

    def __init__(self, agent: ClientAgent):
        self.agent = agent
        self._init_callbacks()  # register callbacks on the given agent

        self.config = Config()
        self.config.workspace = ""  # Reset workspace so that it will computed in start_received
        self.dse        = None
        self.program    = None
        self._stop      = False
        self.sast_report= None
        self._last_id = None
        self._last_id_pc = None
        self._seed_received = set()
        self._probes = []
        self._chkmode = None
        self._seedloc = None
        self._program_slice = None
        self._tracing_enabled = False

        # local attributes for telemetry
        self.nb_to, self.nb_crash = 0, 0
        self._cur_cov_count = 0
        self._last_cov_update = time.time()
        self._seed_queue = queue.Queue()
        self._sending_count = 0
        self.seeds_merged = 0
        self.seeds_rejected = 0

        # Timing stats
        self._start_time = 0
        self._replay_acc = 0

        self.replay_trace_file, self.replay_seed_file = self._initialize_tmp_files()

    def _initialize_tmp_files(self) -> Tuple[Path, Path]:
        ramdisk = Path(self.RAMDISK)
        pid = os.getpid()
        if ramdisk.exists():  # there is a ramdisk available
            dir = ramdisk / f"triton_{pid}"
            dir.mkdir()
            logging.info(f"tmp directory set to: {dir}")
            return dir / self.TMP_TRACE, dir / self.TMP_SEED
        else:
            logging.info(f"tmp directory set to: /tmp")
            return Path(f"/tmp/triton_{pid}.trace"), Path("/tmp/triton_{pid}.seed")

    def add_probe(self, probe: ProbeInterface):
        self._probes.append(probe)

    def _init_callbacks(self):
        self.agent.register_seed_callback(self.seed_received)
        self.agent.register_stop_callback(self.stop_received)

    def init_agent(self, remote: str = "localhost", port: int = 5555):
        self.agent.register_start_callback(self.start_received) # register start because launched manually
        self.agent.connect(remote, port)
        self.agent.start()
        self.agent.send_hello([FuzzingEngineInfo("TRITON", pastisdse.__version__, "pastisttbroker")])

    def start(self):
        self._th = threading.Thread(target=self.run, daemon=True)
        self._th.start()

    def reset(self):
        """ Reset the current DSE to be able to restart from fresh settings """
        self.dse = None  # remove DSE object
        self.config = Config()
        self.config.workspace = ""  # Reset workspace so that it will computed in start_received
        self._last_id_pc = None
        self._last_id = None
        self.sast_report = None
        self._program_slice = None
        self._seed_received = set()
        self.program = None
        self._stop = False
        self._chkmode = None
        self._seedloc = None
        self.nb_to, self.nb_crash = 0, 0
        self._cur_cov_count = 0
        self._last_cov_update = time.time()
        self._tracing_enabled = False
        self._sending_count = 0
        self.seeds_merged = 0
        self.seeds_rejected = 0
        self._start_time = 0
        self._replay_acc = 0
        logging.info("DSE Ready")

    def run(self, online: bool, debug_pp: bool=False):

        # Run while we are not instructed to stop
        while not self._stop:

            if online:  # in offline start_received, seed_received will already have been called
                self.reset()

            # Just wait until the broker says let's go
            while self.dse is None:
                time.sleep(0.10)

            if debug_pp:
                def cb_debug(se, _):
                    se.debug_pp = True
                self.dse.callback_manager.register_pre_execution_callback(cb_debug)

            if not self.run_one(online):
                break

        self.agent.stop()

    def run_one(self, online: bool):
        # Run while we are not instructed to stop
        while not self._stop:

            # wait for seed event
            self._wait_seed_event()
            self._start_time = time.time()

            st = self.dse.explore()

            if not online:
                return False  # in offline whatever the status we stop

            else: # ONLINE
                if st == ExplorationStatus.STOPPED:  # if the exploration stopped just return
                    logging.info("exploration stopped")
                    return False
                elif st == ExplorationStatus.TERMINATED:
                    self.agent.send_stop_coverage_criteria()
                    return True  # Reset and wait for further instruction from the broker
                elif st == ExplorationStatus.IDLE:  # no seed
                    if self._chkmode == CheckMode.ALERT_ONE:
                        self.agent.send_stop_coverage_criteria()  # Warn: the broker we explored the whole search space and did not validated the target
                        return True                               # Make ourself ready to receive a new one
                    else: # wait for seed of peers
                        logging.info("exploration idle (worklist empty)")
                        self.agent.send_log(LogLevel.INFO, "exploration idle (worklist empty)")
                else:
                    logging.error(f"explorator not meant to be in state: {st}")
                    return False

            # Finished an exploration batch
            self.save_stats()  # Save stats

    def _wait_seed_event(self):
        logging.info("wait to receive seeds")
        while not self.dse.seeds_manager.seeds_available() and not self._stop:
            self.try_process_seed_queue()
            time.sleep(0.5)


    def cb_post_execution(self, se: SymbolicExecutor, state: ProcessState):
        """
        This function is called after each execution. In this function we verify
        the ABV_GENERAL alert when a crash occurred during the last execution.

        :param se: The current symbolic executor
        :param state: The current processus state of the execution
        :return: None
        """
        # Send seed that have been executed
        mapper = {SeedStatus.OK_DONE: SeedType.INPUT, SeedStatus.CRASH: SeedType.CRASH, SeedStatus.HANG: SeedType.HANG}
        seed = se.seed
        if seed.status == SeedStatus.NEW:
            logging.warning(f"seed is not meant to be NEW in post execution current:{seed.status.name}")
        elif seed.status in [SeedStatus.CRASH, SeedStatus.HANG]:  # The stats is new send it back again
            if seed not in self._seed_received:  # Do not send back a seed that already came from broker
                self.agent.send_seed(mapper[seed.status], seed.content.files[self.INPUT_FILE_NAME] if seed.is_composite() else seed.content)
        else:  # INPUT
            pass  # Do not send it back again

        # Update some stats
        if se.seed.status == SeedStatus.CRASH:
            self.nb_crash += 1
        elif se.seed.status == SeedStatus.HANG:
            self.nb_to += 1

        # Handle CRASH and ABV_GENERAL
        if se.seed.status == SeedStatus.CRASH and self._last_id:
            alert = self.sast_report.get_alert(self._last_id)
            if alert.type == "ABV_GENERAL":
                logging.info(f'A crash occured with an ABV_GENERAL encountered just before.')
                self.dual_log(LogLevel.INFO, f"Alert [{alert.id}] in {alert.file}:{alert.line}: {alert.type} validation [SUCCESS]")
                alert.validated = True
                self.agent.send_alert_data(AlertData(alert.id, alert.covered, alert.validated, se.seed.content, self._last_id_pc))

        # Process all the seed received
        self.try_process_seed_queue()

        # Print stats
        if self.sast_report:
            cov, va, tot = self.sast_report.get_stats()
            logging.info(f"SAST stats: defaults: [covered:{cov}/{tot}] [validated:{va}/{tot}]")

    def try_process_seed_queue(self):

        while not self._seed_queue.empty() and not self._stop:
            seed, typ = self._seed_queue.get()
            self._process_seed_received(typ, seed)

    def cb_telemetry(self, dse: SymbolicExplorator):
        """
        Callback called after each execution to send telemetry to the broker

        :param dse: SymbolicExplorator
        :return: None
        """
        new_count = dse.coverage.unique_covitem_covered

        if new_count != self._cur_cov_count:         # Coverage has been updated
            self._cur_cov_count = new_count          # update count
            self._last_cov_update = time.time()  # update last coverage update to be now

        if dse.coverage.strategy == CoverageStrategy.PREFIXED_EDGE:
            new_count = len(set(x[1] for x in dse.coverage.covered_items.keys()))  # For prefixed-edge only count edge

        self.agent.send_telemetry(exec_per_sec=int(dse.execution_count / (time.time()-dse.ts)),
                                  total_exec=dse.execution_count,
                                  timeout=self.nb_to,
                                  coverage_block=dse.coverage.unique_instruction_covered,
                                  coverage_edge=new_count if dse.coverage in [CoverageStrategy.EDGE, CoverageStrategy.PREFIXED_EDGE] else 0,
                                  coverage_path=new_count if dse.coverage.strategy == CoverageStrategy.PATH else 0,
                                  last_cov_update=int(self._last_cov_update))

    def cb_on_solving(self, dse: SymbolicExplorator, pstate: ProcessState, edge: Edge, typ: SymExType) -> bool:
        # Only consider conditional and dynamic jumps.
        if typ in [SymExType.SYMBOLIC_READ, SymExType.SYMBOLIC_WRITE]:
            return True

        # Unpack edge.
        src, dst = edge

        # Find the function which holds the basic block of the destination.
        dst_fn = self.program.find_function_from_addr(dst)
        if dst_fn is None:
            logging.warning("Solving edge ({src:#x} -> {dst:#x}) not in a function")
            return True
        else:
            if dst_fn.start in self._program_slice:
                return True
            else:
                logging.info(
                    f"Slicer: reject edge ({src:#x} -> {dst:#x} ({dst_fn.name}) not in slice!")
                return False

    def start_received(self, fname: str, binary: bytes, engine: FuzzingEngineInfo, exmode: ExecMode, fuzzmode: FuzzMode, chkmode: CheckMode,
                       covmode: CoverageMode, seed_inj: SeedInjectLoc, engine_args: str, argv: List[str], sast_report: str=None):
        """
        This function is called when the broker says to start the fuzzing session. Here, we receive all information about
        the program to fuzz and the configuration.

        :param fname: The name of the binary to explore
        :param binary: The content of the binary to explore
        :param engine: The kind of fuzzing engine (should be Triton for this script)
        :param exmode: The mode of the exploration
        :param fuzzmode: The fuzzing mode (instrumented or binary only)
        :param chkmode: The mode of vulnerability check
        :param covmode: The mode of coverage
        :param seed_inj: The location where to inject input
        :param engine_args: The engine arguments
        :param argv: The program arguments
        :param sast_report: The SAST report
        :return: None
        """
        logging.info(f"[BROKER] [START] bin:{fname} engine:{engine.name} exmode:{exmode.name} cov:{covmode.name} chk:{chkmode.name}")

        if self.dse is not None:
            logging.warning("DSE already instanciated (override it)")

        if engine.version != pastisdse.__version__:
            logging.error(f"Pastis-DSE mismatch with one from the server {engine.version} (local: {pastisdse.__version__})")
            return

        self._seedloc = seed_inj

        # ------- Create the TritonDSE configuration file ---------
        if engine_args:
            self.config = Config.from_json(engine_args)
        else:
            self.config = Config()  # Empty configuration
            # Use argv ONLY if no configuration provided
            self.config.program_argv = [f"./{fname}"]
            if argv:
                self.config.program_argv.extend(argv) # Preprend the binary to argv

        """
        Actions taken depending on seed format & co:
        Config    |  Inject  |  Result
        RAW          STDIN      /
        COMPOSITE    STDIN      / (but need 'stdin' in files)
        RAW          ARGV       change to COMPOSITE to be able to inject on argv (and convert seeds on the fly)
        COMPOSITE    ARGV       / (but need 'input_file' in files)
        """
        if seed_inj == SeedInjectLoc.ARGV:  # Make sure we inject input on argv
            if self.config.is_format_raw():
                logging.warning("injection is ARGV thus switch config seed format to COMPOSITE")
                self.config.seed_format = SeedFormat.COMPOSITE
            if "@@" in self.config.program_argv:
                idx = self.config.program_argv.index("@@")
                self.config.program_argv[idx] = self.INPUT_FILE_NAME
            else:
                logging.warning(f"seed inject {self._seedloc.name} but no '@@' found in argv (will likely not work!)")
        else:  # SeedInjectLoc.STDIN
            if engine_args:
                if self.config.is_format_composite():
                    self.dual_log(LogLevel.WARNING, "injecting on STDIN but seed format is COMPOSITE")
            else:  # no config was provided just override
                self.config.seed_format = SeedFormat.RAW
            pass  # nothing to do ?
        # ----------------------------------------------------------

        # If a workspace is given keep it other generate new unique one
        if not self.config.workspace:
            ws = f"/tmp/triton_workspace/{int(time.time())}"
            logging.info(f"Configure workspace to be: {ws}")
            self.config.workspace = ws

        # Create the workspace object in advance (to directly save the binary inside
        workspace = Workspace(self.config.workspace)
        workspace.initialize(flush=False)

        try:
            pkg = BinaryPackage.from_binary(fname, binary, workspace.get_binary_directory())
        except FileNotFoundError:
            logging.error("Invalid package data")
            return

        if pkg.is_quokka():
            logging.info(f"load QuokkaProgram: {pkg.quokka.name}")
            self.program = QuokkaProgram(pkg.quokka, pkg.executable_path)
        else:
            logging.info(f"load Program: {pkg.executable_path.name} [{self._seedloc.name}]")
            program = Program(pkg.executable_path)

            # Make sure the Program is compatible with the local platform
            if self.is_compatible_with_local(program):
                self.program = CleLoader(pkg.executable_path)
            else:
                self.program = program

        if self.program is None:
            self.dual_log(LogLevel.CRITICAL, f"LIEF was not able to parse the binary file {fname}")
            self.agent.stop()
            return

        # Enable local tracing if the binary is compatible with local architecture
        self._tracing_enabled = self.is_compatible_with_local(self.program)
        logging.info(f"Local arch and program arch matching: {self._tracing_enabled}")

        # Update the coverage strategy in the current config (it overrides the config file one)
        try:
            self.config.coverage_strategy = CoverageStrategy(covmode.value)  # names are meant to match
        except Exception as e:
            logging.info(f"Invalid covmode. Not supported by the tritondse library {e}")
            self.agent.stop()
            return

        if sast_report:
            self.sast_report = SASTReport.from_json(sast_report)
            logging.info(f"SAST report loaded: alerts:{len(list(self.sast_report.iter_alerts()))}")

        # Set seed scheduler based on whether tracing is enabled.
        if self._tracing_enabled:
            seed_scheduler_class = WorklistAddressToSet
        else:
            seed_scheduler_class = FreshSeedPrioritizerWorklist

        dse = SymbolicExplorator(self.config, self.program, workspace=workspace, seed_scheduler_class=seed_scheduler_class)

        # Register common callbacks
        dse.callback_manager.register_new_input_callback(self.send_seed_to_broker) # must be the second cb
        dse.callback_manager.register_post_execution_callback(self.cb_post_execution)
        dse.callback_manager.register_exploration_step_callback(self.cb_telemetry)

        for probe in self._probes:
            dse.callback_manager.register_probe(probe)

        # Save check mode
        self._chkmode = chkmode

        if chkmode == CheckMode.CHECK_ALL:
           dse.callback_manager.register_probe(UAFSanitizer())
           dse.callback_manager.register_probe(NullDerefSanitizer())
           dse.callback_manager.register_probe(FormatStringSanitizer())
           #dse.callback_manager.register_probe(IntegerOverflowSanitizer())
           # TODO Buffer overflow

        elif chkmode == CheckMode.ALERT_ONLY:
           dse.callback_manager.register_function_callback('__sast_alert_placeholder', self.intrinsic_callback)

        elif chkmode == CheckMode.ALERT_ONE:  # targeted approach
            if not isinstance(self.program, QuokkaProgram):
                logging.error(f"Targeted mode [{chkmode.name}] requires a Quokka program")
                self.agent.stop()
                return

            target_addr = self.config.custom['target']  # retrieve the target address to reach
            dse.callback_manager.register_post_addr_callback(target_addr, self.intrinsic_callback)

            # NOTE Target address must be the starting address of a basic block.
            slice_from = self.program.find_function_addr('main')
            slice_to = target_addr

            if slice_from and slice_to:
                # Find the functions that correspond to the from and to addresses.
                slice_from_fn = self.program.find_function_from_addr(slice_from)
                slice_to_fn = self.program.find_function_from_addr(slice_to)
                logging.info(f"launching exploration in targeted mode on: 0x{target_addr:08x} in {slice_to_fn.name}")

                if slice_from_fn and slice_to_fn:
                    # NOTE Generate call graph with backedges so when we do the
                    #      slice it also includes functions that are called in
                    #      the path from the source to the destination of the
                    #      slice.
                    call_graph = self.program.get_call_graph(backedge_on_ret=True)

                    logging.info(f'Slicing program from {slice_from:#x} ({slice_from_fn.name}) to {slice_to:#x} ({slice_to_fn.name})')

                    self._program_slice = QuokkaProgram.get_slice(call_graph, slice_from_fn.start, slice_to_fn.start)

                    logging.info(f'Call graph (full): #nodes: {len(call_graph.nodes)}, #edges: {len(call_graph.edges)}')
                    logging.info(f'Call graph (sliced): #nodes: {len(self._program_slice.nodes)}, #edges: {len(self._program_slice.edges)}')

                    dse.callback_manager.register_on_solving_callback(self.cb_on_solving)
                else:
                    logging.warning(f'Invalid source or target function, skipping slicing!')
            else:
                logging.warning(f'Invalid source or target addresses, skipping slicing!')

        # will trigger the dse to start has another thread is waiting for self.dse to be not None
        self.dse = dse

    def _get_seed(self, raw_seed: bytes) -> Seed:
        # Convert seed to CompositeData
        seed = Seed.from_bytes(raw_seed)

        if self.config.is_format_composite() and seed.is_raw() and self._seedloc == SeedInjectLoc.ARGV:
            logging.debug("convert raw seed to composite")
            return Seed(CompositeData(files={self.INPUT_FILE_NAME: seed.content}))

        elif self.config.is_format_composite() and seed.is_raw() and self._seedloc == SeedInjectLoc.STDIN:
            logging.warning("Seed is RAW but format is COMPOSITE with injection on STDIN")
            return Seed(CompositeData(files={"stdin": seed.content}))

        else:
            return seed

    def seed_received(self, typ: SeedType, seed: bytes):
        """
        This function is called when we receive a seed from the broker.

        :param typ: The type of the seed
        :param seed: The seed
        :return: None
        """
        seed = self._get_seed(seed)

        if seed in self._seed_received:
            logging.warning(f"receiving seed already known: {seed.hash} (dropped)")
            return
        else:
            self._seed_queue.put((seed, typ))
            logging.info(f"seed received {seed.hash} (pool: {self._seed_queue.qsize()})")


    def _process_seed_received(self, typ: SeedType, seed: Seed):
        """
        This function is called when we receive a seed from the broker.

        :param typ: The type of the seed
        :param seed: The seed
        :return: None
        """
        try:
            if not self._tracing_enabled:
                # Accept all seeds
                self.dse.add_input_seed(seed)

            else:  # Try running the seed to know whether to keep it
                # NOTE: re-run the seed regardless of its status
                coverage = None
                logging.info(f"process seed received {seed.hash} (pool: {self._seed_queue.qsize()})")

                data = seed.content.files[self.INPUT_FILE_NAME] if seed.is_composite() else seed.bytes()
                self.replay_seed_file.write_bytes(data)
                # Adjust injection location before calling QBDITrace
                if self._seedloc == SeedInjectLoc.STDIN:
                    stdin_file = str(self.replay_seed_file)
                    argv = self.config.program_argv
                else:  # SeedInjectLoc.ARGV
                    stdin_file = None
                    try:
                        # Replace 'input_file' in argv with the temporary file name created
                        argv = self.config.program_argv[:]
                        idx = argv.index(self.INPUT_FILE_NAME)
                        argv[idx] = str(self.replay_seed_file)
                    except ValueError:
                        logging.error(f"seed injection {self._seedloc.name} but can't find 'input_file' on program argv")
                        return

                try:
                    # Run the seed and determine whether it improves our current coverage.
                    t0 = time.time()
                    if QBDITrace.run(self.config.coverage_strategy,
                                          str(self.program.path.resolve()),
                                          argv[1:] if len(argv) > 1 else [],
                                          output_path=str(self.replay_trace_file),
                                          stdin_file=stdin_file,
                                          cwd=Path(self.program.path).parent,
                                          timeout=60):
                        coverage = QBDITrace.from_file(str(self.replay_trace_file)).coverage
                    else:
                        logging.warning("Cannot load the coverage file generated (maybe had crashed?)")
                        coverage = None
                    self._replay_acc += time.time() - t0  # Save time spent replaying inputs
                except FileNotFoundError:
                    logging.warning("Cannot load the coverage file generated (maybe had crashed?)")
                except TraceException:
                    logging.warning('There was an error while trying to re-run the seed')

                if not coverage:
                    logging.warning(f"coverage not found after replaying: {seed.hash} [{typ.name}] (add it anyway)")
                    # Add the seed anyway, if it was not possible to re-run the seed.
                    # TODO Set seed.coverage_objectives as "empty" (use ellipsis
                    # object). Modify WorklistAddressToSet to support it.
                    self.seeds_merged += 1
                    self.dse.add_input_seed(seed)
                else:
                    # Check whether the seed improves the current coverage.
                    if self.dse.coverage.improve_coverage(coverage):
                        logging.info(f"seed added {seed.hash} [{typ.name}] (coverage merged)")
                        self.seeds_merged += 1
                        self.dse.coverage.merge(coverage)
                        self.dse.seeds_manager.worklist.update_worklist(coverage)

                        seed.coverage_objectives = self.dse.coverage.new_items_to_cover(coverage)
                        self.dse.add_input_seed(seed)
                    else:
                        logging.info(f"seed archived {seed.hash} [{typ.name}] (NOT merging coverage)")
                        self.seeds_rejected += 1
                        #self.dse.seeds_manager.archive_seed(seed)
                        # logging.info(f"seed archived {seed.hash} [{typ.name}]")

            self._seed_received.add(seed)  # Remember seed received not to send them back
        except FileNotFoundError as e:
            # NOTE If reset() is call during the execution of this function,
            #      self.dse will be set to None and an AttributeError will occur.
            logging.warning(f"receiving seeds while the DSE is not instantiated {e}")

        rcv = len(self._seed_received)
        logging.info(f"seeds recv: {rcv} | merged {self.seeds_merged} [{(self.seeds_merged/rcv) * 100:.2f}%]"
                     f" rejected {self.seeds_rejected} [{(self.seeds_rejected/rcv) * 100:.2f}%]")

    def stop_received(self):
        """
        This function is called when the broker says stop. (Called from the agent thread)
        """
        logging.info(f"[BROKER] [STOP]")

        if self.dse:
            self.dse.stop_exploration()

        self.save_stats()  # Save stats

        self._stop = True
        # self.agent.stop()  # Can't call it here as this function executed from within agent thread

    def save_stats(self):
        stat_file = self.dse.workspace.get_metadata_file_path(self.STAT_FILE)
        data = {
            "total_time": time.time() - self._start_time,
            "emulation_time": self.dse.total_emulation_time,  # Note: includes replay time but not solving
            "solving_time": self.dse.seeds_manager.total_solving_time,
            "replay_time": self._replay_acc,
            "seed_accepted": self.seeds_merged,
            "seed_rejected": self.seeds_rejected,
            "seed_received": self.seeds_merged + self.seeds_rejected
        }
        stat_file.write_text(json.dumps(data))

    def dual_log(self, level: LogLevel, message: str) -> None:
        """
        Helper function to log message both in the local log system and also
        to the broker.

        :param level: LogLevel message type
        :param message: string message to log
        :return: None
        """
        mapper = {LogLevel.DEBUG: "debug",
                  LogLevel.INFO: "info",
                  LogLevel.CRITICAL: "critical",
                  LogLevel.WARNING: "warning",
                  LogLevel.ERROR: "error"}
        log_f = getattr(logging, mapper[level])
        log_f(message)
        self.agent.send_log(level, message)

    def send_seed_to_broker(self, se: SymbolicExecutor, state: ProcessState, seed: Seed):
        if seed not in self._seed_received:  # Do not send back a seed that already came from broker
            self._sending_count += 1
            logging.info(f"Sending new: {seed.hash} [{self._sending_count}]")
            bytes = seed.content.files[self.INPUT_FILE_NAME] if seed.is_composite() else seed.content
            self.agent.send_seed(SeedType.INPUT, bytes)

    def intrinsic_callback(self, se: SymbolicExecutor, state: ProcessState, addr: Addr):
        """
        This function is called when an intrinsic call occurs in order to verify
        defaults and vulnerabilities.

        :param se: The current symbolic executor
        :param state: The current processus state of the execution
        :param addr: The instruction address of the intrinsic call
        :return: None
        """
        alert_id = state.get_argument_value(0)
        self._last_id = alert_id
        self._last_id_pc = se.previous_pc

        def status_changed(a, cov, vld):
            return a.covered != cov or a.validated != vld

        if self.sast_report:
            # Retrieve the SASTAlert object from the report
            try:
                alert = self.sast_report.get_alert(alert_id)
                cov, vld = alert.covered, alert.validated
            except IndexError:
                logging.warning(f"Intrinsic id {alert_id} not found in report (ignored)")
                return

            if not alert.covered:
                self.dual_log(LogLevel.INFO, f"Alert [{alert.id}] in {alert.file}:{alert.line}: {alert.type} covered !")
                alert.covered = True  # that might also set validated to true!

            if not alert.validated:  # If of type VULNERABILITY and not yet validated
                res = self.check_alert_dispatcher(alert.code, se, state, addr)
                if res:
                    alert.validated = True
                    self.dual_log(LogLevel.INFO, f"Alert [{alert.id}] in {alert.file}:{alert.line}: {alert.type} validation [SUCCESS]")
                    if se.seed.is_status_set():
                        logging.warning(f"Status already set ({se.seed.status}) for seed {se.seed.hash} (override with CRASH)")
                    se.seed.status = SeedStatus.CRASH  # Mark the seed as crash, as it validates an alert
                else:
                    logging.info(f"Alert [{alert.id}] in {alert.file}:{alert.line}: validation [FAIL]")

            if status_changed(alert, cov, vld):  # If either coverage or validation were improved print stats
                # Send updates to the broker
                self.agent.send_alert_data(AlertData(alert.id, alert.covered, alert.validated, se.seed.content, se.previous_pc))
                cov, vals, tot = self.sast_report.get_stats()
                logging.info(f"SAST stats: defaults: [covered:{cov}/{tot}] [validated:{vals}/{tot}]")

                if self.sast_report.all_alerts_validated() or (self._chkmode == CheckMode.ALERT_ONE and alert.validated):
                    self._do_stop_all_alerts_validated()

        else:  # Kind of autonomous mode. Try to check it even it is not bound to a report
            # Retrieve alert type from parameters
            alert_type = se.pstate.get_string_argument(1)
            try:
                if self.check_alert_dispatcher(alert_type, se, state, addr):
                    logging.info(f"Alert {alert_id} of type {alert_type} [VALIDATED]")
                else:
                    logging.info(f"Alert {alert_id} of type {alert_type} [NOT VALIDATED]")
            except KeyError:
                logging.error(f"Alert type {alert_type} not recognized")


    def check_alert_dispatcher(self, type: str, se: SymbolicExecutor, state: ProcessState, addr: Addr) -> bool:
        """
        This function is called by intrinsic_callback in order to verify defaults
        and vulnerabilities.

        :param type: Type of the alert as a string
        :param se: The current symbolic executor
        :param state: The current processus state of the execution
        :param addr: The instruction address of the intrinsic call
        :return: True if a vulnerability has been verified
        """
        # BUFFER_OVERFLOW related alerts
        if type == "SV_STRBO_UNBOUND_COPY":
            size = se.pstate.get_argument_value(2)
            ptr = se.pstate.get_argument_value(3)

            # Runtime check
            if len(se.pstate.get_memory_string(ptr)) >= size:
                # FIXME: Do we have to define the seed as CRASH even if there is no crash?
                # FIXME: Maybe we have to define a new TAG like BUG or VULN or whatever
                return True

            # Symbolic check
            actx = se.pstate.actx
            predicate = [se.pstate.tt_ctx.getPathPredicate()]

            # For each memory cell, try to proof that they can be different from \0
            for i in range(size + 1): # +1 in order to proof that we can at least do an off-by-one
                cell = se.pstate.tt_ctx.getMemoryAst(MemoryAccess(ptr + i, CPUSIZE.BYTE))
                predicate.append(cell != 0)

            # FIXME: Maybe we can generate models until unsat in order to find the bigger string

            model = se.pstate.tt_ctx.getModel(actx.land(predicate))
            if model:
                crash_seed = mk_new_crashing_seed(se, model)
                se.workspace.save_seed(crash_seed)
                logging.info(f'Model found for a seed which may lead to a crash ({crash_seed.filename})')
                return True

            return False

        ######################################################################

        # BUFFER_OVERFLOW related alerts
        elif type == "SV_STRBO_BOUND_COPY_OVERFLOW":
            dst_size = se.pstate.get_argument_value(2)
            ptr_inpt = se.pstate.get_argument_value(3)
            max_size = se.pstate.get_argument_value(4)

            # Runtime check
            if max_size >= dst_size and len(se.pstate.get_memory_string(ptr_inpt)) >= dst_size:
                # FIXME: Do we have to define the seed as CRASH even if there is no crash?
                # FIXME: Maybe we have to define a new TAG like BUG or VULN or whatever
                return True

            # Symbolic check
            actx = se.pstate.actx
            max_size_s = se.pstate.get_argument_symbolic(4).getAst()
            predicate = [se.pstate.tt_ctx.getPathPredicate(), max_size_s >= dst_size]

            # For each memory cell, try to proof that they can be different from \0
            for i in range(dst_size + 1): # +1 in order to proof that we can at least do an off-by-one
                cell = se.pstate.tt_ctx.getMemoryAst(MemoryAccess(ptr_inpt + i, CPUSIZE.BYTE))
                predicate.append(cell != 0)

            # FIXME: Maybe we can generate models until unsat in order to find the bigger string

            model = se.pstate.tt_ctx.getModel(actx.land(predicate))
            if model:
                crash_seed = mk_new_crashing_seed(se, model)
                se.workspace.save_seed(crash_seed)
                logging.info(f'Model found for a seed which may lead to a crash ({crash_seed.filename})')
                return True

            return False

        ######################################################################

        # BUFFER_OVERFLOW related alerts
        elif type == "ABV_GENERAL":
            logging.warning(f'ABV_GENERAL encounter but can not check the issue. This issue will be handle if the program will crash.')
            return False

        ######################################################################

        # All INTEGER_OVERFLOW related alerts
        elif type == "NUM_OVERFLOW":
            return IntegerOverflowSanitizer.check(se, state, state.current_instruction)

        ######################################################################

        # All USE_AFTER_FREE related alerts
        elif type in ["UFM_DEREF_MIGHT", "UFM_FFM_MUST", "UFM_FFM_MIGHT"]:
            ptr = se.pstate.get_argument_value(2)
            return UAFSanitizer.check(se, state, ptr, f'UAF detected at {ptr:#x}')

        ######################################################################

        # All FORMAT_STRING related alerts
        elif type in ["SV_TAINTED_FMTSTR", "SV_FMTSTR_GENERIC"]:
            ptr = se.pstate.get_argument_value(2)
            return FormatStringSanitizer.check(se, state, addr, ptr)

        ######################################################################

        # All INVALID_MEMORY related alerts
        # FIXME: NPD_CHECK_MIGHT and NPD_CONST_CALL are not supported by klocwork-alert-inserter
        elif type in ["NPD_FUNC_MUST", "NPD_FUNC_MIGHT", "NPD_CHECK_MIGHT", "NPD_CONST_CALL"]:
            ptr = se.pstate.get_argument_value(2)
            return NullDerefSanitizer.check(se, state, ptr, f'Invalid memory access at {ptr:#x}')

        ######################################################################

        elif type == "MISRA_ETYPE_CATEGORY_DIFFERENT_2012":
            expr = se.pstate.get_argument_symbolic(2).getAst()

            # Runtime check
            if expr.isSigned():
                # FIXME: Do we have to define the seed as CRASH even if there is no crash?
                # FIXME: Maybe we have to define a new TAG like BUG or VULN or whatever
                return True

            # Symbolic check
            actx = se.pstate.tt_ctx.getAstContext()
            size = expr.getBitvectorSize() - 1
            predicate = [se.pstate.tt_ctx.getPathPredicate(), actx.extract(size - 1, size - 1, expr) == 1]

            model = se.pstate.tt_ctx.getModel(actx.land(predicate))
            if model:
                crash_seed = mk_new_crashing_seed(se, model)
                se.workspace.save_seed(crash_seed)
                logging.info(f'Model found for a seed which may lead to a crash ({crash_seed.filename})')
                return True
            return False

        else:
            logging.error(f"Unsupported alert kind {type}")


    def _do_stop_all_alerts_validated(self) -> None:
        """
        Function called if all alerts have been covered and validated. All data are meant to
        have been transmitted to the broker, but writes down locally the CSV anyway
        :return: None
        """
        logging.info("All defaults and vulnerability have been covered !")

        # Write the final CSV in the workspace directory
        out_file = self.dse.workspace.get_metadata_file_path("klocwork_coverage_results.csv")
        self.sast_report.write_csv(out_file)

        # Stop the dse exploration
        self.dse.terminate_exploration()


    def is_compatible_with_local(self, program: Program) -> bool:
        """
        Checks whether the given program is compatible with the current architecture
        and platform.

        :param program: Program
        :return: True if the program can be run locally
        """
        arch_m = {"i386": Architecture.X86, "x86_64": Architecture.X86_64, "armv7l": Architecture.ARM32, "aarch64": Architecture.AARCH64}
        plfm_m = {"Linux": Platform.LINUX, "Windows": Platform.WINDOWS, "MacOS": Platform.MACOS, "iOS": Platform.IOS}
        local_arch, local_plfm = arch_m[platform.machine()], plfm_m[platform.system()]
        return program.architecture == local_arch and program.platform == local_plfm



# builtin imports
import hashlib
import logging
import stat
import threading
import time

from pathlib import Path
from typing import List, Union

# Third party imports
from libpastis import ClientAgent, BinaryPackage, SASTReport
from libpastis.types import CheckMode, CoverageMode, ExecMode, FuzzingEngineInfo, SeedInjectLoc, SeedType, State, \
                            LogLevel, AlertData, FuzzMode


# Local imports
import pastisaflpp
from pastisaflpp.replay import Replay
from pastisaflpp.aflpp import AFLPPProcess
from pastisaflpp.workspace import Workspace


# Inotify logs are very talkative, set them to ERROR
for logger in (logging.getLogger(x) for x in ["watchdog.observers.inotify_buffer", 'watchdog.observers', "watchdog"]):
    logger.setLevel(logging.ERROR)


class AFLPPDriver:

    def __init__(self, agent: ClientAgent, telemetry_frequency: int = 30):

        # Internal objects
        self._agent = agent
        self.workspace = Workspace()
        self.aflpp = AFLPPProcess()

        # Parameters received through start_received
        self.__exec_mode = None   # SINGLE_RUN, PERSISTENT
        self.__check_mode = None  # CHECK_ALL, ALERT_ONLY
        self.__seed_inj = None    # STDIN or ARGV
        self.__report = None      # Klocwork report if supported

        # Target data
        self.__package = None
        self.__target_args = None  # Kept for replay

        self.__setup_agent()

        # Configure hookds on workspace
        self.workspace.add_creation_hook(self.workspace.corpus_dir, self.__send_seed)
        self.workspace.add_creation_hook(self.workspace.crash_dir, self.__send_crash)
        self.workspace.add_file_modification_hook(self.workspace.stats_dir, self.__send_telemetry)

        # Telemetry frequency
        self._tel_frequency = telemetry_frequency
        self._tel_last = time.time()

        # Runtime data
        self._tot_seeds = 0
        self._seed_recvs = set()  # Seed received to make sure NOT to send them back

        # Variables for replay
        self._replay_thread = None
        self._queue_to_send = []
        self._started = False

    @staticmethod
    def hash_seed(seed: bytes):
        return hashlib.md5(seed).hexdigest()

    def start(self, package: BinaryPackage, argv: List[str], exmode: ExecMode, fuzzmode: FuzzMode, seed_inj: SeedInjectLoc, engine_args: str):
        # Write target to disk.
        self.__package = package
        self.__target_args = argv

        self.workspace.start()  # Start looking at directories

        logging.info(f"Start process (injectloc: {seed_inj.name})")
        self.aflpp.start(str(package.executable_path.absolute()),
                         argv,
                         self.workspace,
                         exmode,
                         fuzzmode,
                         seed_inj == SeedInjectLoc.STDIN,
                         engine_args,
                         str(package.cmplog.absolute()) if package.cmplog else None,
                         str(package.dictionary.absolute()) if package.dictionary else None)
        self._started = True

        # Start the replay worker (note that the queue might already have started to be filled by agent thread)
        self._replay_thread = threading.Thread(target=self.replay_worker, daemon=True)
        self._replay_thread.start()

    def stop(self):
        self.aflpp.stop()
        self.workspace.stop()
        self._started = False  # should stop the replay thread

    def replay_worker(self):
        while True:
            if not self._started:
                break  # Break when the fuzzer stops
            if self._queue_to_send:
                filename, res = self._queue_to_send.pop(0)
                if not self.__check_seed_alert(filename, is_crash=res):
                    break
            time.sleep(0.05)

    @property
    def started(self):
        return self._started

    def add_seed(self, seed: bytes):
        seed_path = self.workspace.dynamic_input_dir / f"seed-{hashlib.md5(seed).hexdigest()}"
        seed_path.write_bytes(seed)

    def init_agent(self, remote: str = "localhost", port: int = 5555):
        self._agent.register_start_callback(self.start_received)  # Register start because launched manually (not by pastisd)
        self._agent.connect(remote, port)
        self._agent.start()
        # Send initial HELLO message, whick will make the Broker send the START message.
        self._agent.send_hello([FuzzingEngineInfo("AFLPP", pastisaflpp.__version__, "aflppbroker")])

    def run(self):
        self.aflpp.wait()

    def __setup_agent(self):
        # Register callbacks.
        self._agent.register_seed_callback(self.__seed_received)
        self._agent.register_stop_callback(self.__stop_received)

    def __send_seed(self, filename: Path):
        self.__send(filename, SeedType.INPUT)

    def __send_crash(self, filename: Path):
        # Skip README file that AFL adds to the crash folder.
        if filename.name != 'README.txt':
            self.__send(filename, SeedType.CRASH)

    def __send(self, filename: Path, typ: SeedType):
        self._tot_seeds += 1
        file = Path(filename)
        raw = file.read_bytes()
        h = self.hash_seed(raw)
        logging.debug(f'[{typ.name}] Sending new: {h} [{self._tot_seeds}]')
        if h not in self._seed_recvs:
            self._agent.send_seed(typ, raw)
        else:
            logging.info("seed (previously sent) do not send it back")
        self._queue_to_send.append((filename, True if typ == SeedType.CRASH else False))

    def __check_seed_alert(self, filename: Path, is_crash: bool) -> bool:
        p = Path(filename)
        # Only rerun the seed if in alert only mode and a SAST report was provided
        if self.__check_mode == CheckMode.ALERT_ONLY and self.__report:

            # Rerun the program with the seed
            run = Replay.run(self.__package.executable_path.absolute(), self.__target_args, stdin_file=filename, timeout=5, cwd=str(self.workspace.target_dir))

            # FIXME: Do same checks for AFL++ LOOP stuff for persistency mode
            # if run.is_hf_iter_crash():
            #     self.dual_log(LogLevel.ERROR, f"Disable replay engine (because code uses HF_ITER)")
            #     return False

            # Iterate all covered alerts
            for id in run.alert_covered:
                alert = self.__report.get_alert(id)
                if not alert.covered:
                    alert.covered = True
                    logging.info(f"New alert covered {alert} [{alert.id}]")
                    self._agent.send_alert_data(AlertData(alert.id, alert.covered, False, p.read_bytes()))

            # Check if the target has crashed and if so tell the broker which one
            if run.has_crashed() or run.is_asan_without_crash():  # Also consider ASAN warning as detection
                if not run.crashing_id:
                    self.dual_log(LogLevel.WARNING, f"Crash on {filename.name} but can't link it to a Klocwork alert (maybe bonus !)")
                else:
                    alert = self.__report.get_alert(run.crashing_id)
                    if not alert.validated:
                        alert.validated = True
                        bugt, aline = run.asan_info()
                        self.dual_log(LogLevel.INFO, f"AFLPP new alert validated {alert} [{alert.id}] ({aline})  (asan no crash: {run.is_asan_without_crash()})")
                        self._agent.send_alert_data(AlertData(alert.id, alert.covered, alert.validated, p.read_bytes()))
            else:
                if is_crash:
                    self.dual_log(LogLevel.WARNING, f"crash not reproducible by rerunning seed: {filename.name}")

            if run.has_hanged():  # AFLPP does not stores 'hangs' it will have been sent as corpus or crash
                self.dual_log(LogLevel.WARNING, f"Seed {filename} was hanging in replay")
        return True

    def __send_telemetry(self, filename: Path):
        if filename.name != AFLPPProcess.STAT_FILE:
            return

        now = time.time()
        if now < (self._tel_last + self._tel_frequency):
            return
        self._tel_last = now

        logging.debug(f'[TELEMETRY] Stats file updated: {filename}')

        with open(filename, 'r') as stats_file:
            try:
                stats = {}
                for line in stats_file.readlines():
                    k, v = line.strip('\n').split(':')
                    stats[k.strip()] = v.strip()

                state = State.RUNNING
                last_cov_update = int(stats['last_update'])
                total_exec = int(stats['execs_done'])
                exec_per_sec = int(float(stats['execs_per_sec']))
                timeout = int(stats['unique_hangs']) if 'unique_hangs' in stats else None # N/A in AFL-QEMU.
                coverage_edge = int(stats['total_edges'])
                cycle = int(stats['cycles_done'])
                coverage_path = int(stats['paths_total']) if 'paths_total' in stats else None # N/A in AFL-QEMU.

                # NOTE: `coverage_block` does not apply for AFLPP.
                self._agent.send_telemetry(state=state,
                                           exec_per_sec=exec_per_sec,
                                           total_exec=total_exec,
                                           cycle=cycle,
                                           timeout=timeout,
                                           coverage_edge=coverage_edge,
                                           coverage_path=coverage_path,
                                           last_cov_update=last_cov_update)
            except:
                logging.error(f'Error retrieving stats!')

    def start_received(self, fname: str, binary: bytes, engine: FuzzingEngineInfo, exmode: ExecMode, fuzzmode: FuzzMode, chkmode: CheckMode,
                       _: CoverageMode, seed_inj: SeedInjectLoc, engine_args: str, argv: List[str], sast_report: str = None):
        logging.info(f"[START] bin:{fname} engine:{engine.name} exmode:{exmode.name} seedloc:{seed_inj.name} chk:{chkmode.name}")
        if self.started:
            self._agent.send_log(LogLevel.CRITICAL, "Instance already started!")
            return

        if engine.name != "AFLPP":
            logging.error(f"Wrong fuzzing engine received {engine.name} while I am Honggfuzz")
            self._agent.send_log(LogLevel.ERROR, f"Invalid fuzzing engine received {engine.name} can't do anything")
            return
        if engine.version != pastisaflpp.__version__:
            logging.error(f"Wrong fuzzing engine version {engine.version} received")
            self._agent.send_log(LogLevel.ERROR, f"Invalid fuzzing engine version {engine.version} do nothing")
            return

        # Retrieve package out of the binary received
        try:
            package = BinaryPackage.from_binary(fname, binary, self.workspace.target_dir)
        except FileNotFoundError:
            logging.error("Invalid package received")
            return
        except ValueError:
            logging.error("Invalid package received")
            return

        if sast_report:
            logging.info("Loading SAST report")
            self.__report = SASTReport.from_json(sast_report)

        self.__check_mode = chkmode  # CHECK_ALL, ALERT_ONLY

        self.start(package, argv, exmode, fuzzmode, seed_inj, engine_args)

    def __seed_received(self, typ: SeedType, seed: bytes):
        h = self.hash_seed(seed)
        logging.info(f"[SEED] received  {h} ({typ.name})")
        self._seed_recvs.add(h)
        self.add_seed(seed)

    def __stop_received(self):
        logging.info(f"[STOP]")

        self.stop()

    def dual_log(self, level: LogLevel, message: str) -> None:
        """
        Helper function to log message both in the local log system and also
        to the broker.

        :param level: LogLevel message type
        :param message: string message to log
        :return: None
        """
        mapper = {LogLevel.DEBUG: "debug",
                  LogLevel.INFO: "info",
                  LogLevel.CRITICAL: "critical",
                  LogLevel.WARNING: "warning",
                  LogLevel.ERROR: "error"}
        log_f = getattr(logging, mapper[level])
        log_f(message)
        self._agent.send_log(level, message)

    def add_initial_seed(self, file: Union[str, Path]):
        p = Path(file)
        logging.info(f"add initial seed {file.name}")
        # Write seed to disk.
        seed_path = self.workspace.input_dir / p.name
        seed_path.write_bytes(p.read_bytes())

    def aflpp_available(self):
        return self.aflpp.aflpp_environ_check()



# built-in imports
import logging
from pathlib import Path
from typing import Union, List, Dict
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import json
from datetime import timedelta
from hashlib import md5
import base64
from collections import Counter

# third-party imports
from rich.console import Console
from rich.table import Table

# local imports
from pastisbenchmark.results import InputCovDelta, CampaignResult
from pastisbenchmark.models import CampaignStats, InputEntry, CoverageEntry, ExecEntry, SeedSharingEntry, SmtEntry
from tritondse import CoverageStrategy, SmtSolver, BranchSolvingStrategy


class Plotter(object):

    LABEL_SZ = 18
    TICK_SZ = 13
    FONT_SZ = 18
    LEGEND_SZ = 8

    PLOT_DIR = "plots"

    def __init__(self, name: str, timeout: int):
        self.fig, (self.ax1, self.ax2) = plt.subplots(1, 2)
        self.name = name
        self._timeout = timeout

        # self._configure_plot(self.ax1, ylabel="coverage (edge)")
        # self._configure_plot(self.ax2, ylabel="coverage (edge)", is_log=True)

    def _configure_plot(self, plot, ylabel: str, is_log: bool = False):
        plot.tick_params(axis='both', which='major', labelsize=self.TICK_SZ)
        plot.tick_params(axis='both', which='minor', labelsize=self.TICK_SZ)
        plot.set_title(f"{self.name} {'(logscale)' if is_log else ''}", fontsize=self.FONT_SZ)
        plot.set(xlabel='seconds', ylabel=ylabel)
        plot.yaxis.label.set_size(self.LABEL_SZ)
        plot.xaxis.label.set_size(self.LABEL_SZ)
        plot.legend(prop={'size': self.LEGEND_SZ})
        if is_log:
            plot.set_xscale("log")

    def add_campaign_to_plot(self, campaign: CampaignResult, show_union: bool=True):
        """ Iterate all stat_items and generate coverage plot."""
        for fuzzer, results in campaign.results:
            is_all_fuzzer = bool(fuzzer == CampaignResult.ALL_FUZZER)
            if fuzzer == CampaignResult.SEED_FUZZER:
                continue
            if is_all_fuzzer and campaign.is_half_duplex and not show_union:
                continue
            if campaign.is_full_duplex and not is_all_fuzzer:  # Only print the ALL fuzzer in fullduplex
                continue
            name = self.format_fuzzer_name(campaign, fuzzer)
            # fmt = self.format_plot(fuzzer, campaign.is_full_duplex)

            marker = "--" if campaign.is_half_duplex and is_all_fuzzer else "-"
            color = self.format_plot(campaign, fuzzer)

            self.add_to_plot(self.ax1, name, results, is_all_fuzzer, linestyle=marker, color=color)
            self.add_to_plot(self.ax2, name, results, is_all_fuzzer, linestyle=marker, color=color)

    def add_to_plot(self, plot, fuzzer: str, results: List[InputCovDelta], use_global: bool, **kwargs):
        xaxe = [x.time_elapsed for x in results]
        yaxe = [(x.overall_coverage_sum if use_global else x.fuzzer_coverage_sum) for x in results]

        if not yaxe:
            print(f"no plot for {fuzzer}")
            return

        # Add dummy value to make horizontal line
        xaxe.append(self._timeout)
        yaxe.append(yaxe[-1])

        plot.plot(xaxe, yaxe, label=fuzzer, linewidth=2, **kwargs)

    def format_fuzzer_name(self, campaign: CampaignResult, fuzzer: str, short: bool=False) -> str:
        if fuzzer == CampaignResult.ALL_FUZZER:
            return campaign.slug_name
        elif fuzzer == CampaignResult.SEED_FUZZER:
            return CampaignResult.SEED_FUZZER
        elif "TT" in fuzzer:
            if short:
                return "TritonDSE"
            config = campaign.fuzzers_config[fuzzer]
            cov_name = {CoverageStrategy.BLOCK: "B", CoverageStrategy.EDGE: "E", CoverageStrategy.PREFIXED_EDGE: "PE",
                        CoverageStrategy.PATH: "P"}[config.coverage_strategy]
            param = [
                "R" if BranchSolvingStrategy.COVER_SYM_READ in config.branch_solving_strategy else "-",
                "W" if BranchSolvingStrategy.COVER_SYM_WRITE in config.branch_solving_strategy else "-",
                "J" if BranchSolvingStrategy.COVER_SYM_DYNJUMP in config.branch_solving_strategy else "-"
            ]
            solver = {SmtSolver.Z3: "Z3", SmtSolver.BITWUZLA: "BZLA"}[config.smt_solver]
            return f"TritonDSE[{cov_name}][{''.join(param)}][{solver}]"
        elif "AFLPP" in fuzzer:
            return "AFL++"
        elif "HF" in fuzzer:
            return "Honggfuzz"
        else:
            return fuzzer

    def format_plot(self, campaign, fuzzer) -> str:
        green = "#30a230"
        grey = "#1f77b4"
        brown = "#944b0c"
        if fuzzer == CampaignResult.ALL_FUZZER:
            if campaign.is_full_duplex:
                return green if campaign.has_honggfuzz() else brown
            else:
                return grey
        elif "TT" in fuzzer:
            return "#d62728"
        elif "AFLPP" in fuzzer:
            return "#ff8214"
        elif "HF" in fuzzer:
            return "#9467bd"
        else:
            return ""


    def add_triton_input(self, campaign: CampaignResult):
        if campaign.is_full_duplex:
            results = campaign.fuzzers_items[campaign.ALL_FUZZER]
            X = [x.time_elapsed for x in results if "TT" in x.fuzzer]
            Y = [x.overall_coverage_sum for x in results if "TT" in x.fuzzer]
            self.ax1.plot(X, Y, 'b.', label="TT input")
            self.ax2.plot(X, Y, 'b.', label="TT input")
        else:
            logging.warning(f"campaign:{campaign.workspace.root} not full duplex do not show triton inputs")

    def show(self):
        self._configure_plot(self.ax1, ylabel="coverage (edge)")
        self._configure_plot(self.ax2, ylabel="coverage (edge)", is_log=True)
        plt.show()

    def save_to(self, dir: Union[str, Path]) -> None:
        plt.savefig(dir / "plot.pdf", dpi=600)


    def calculate_stats(self, campaign: CampaignResult) -> CampaignStats:
        input_stats = self._calcul_input_stats(campaign)
        coverage_stats = self._calcul_coverage_stats(campaign)
        exec_stats = self._calcul_exec_stats(campaign)
        seed_sharing_stats = self._calcul_seed_sharing_stats(campaign)
        smt_stats = self._calcul_smt_stats(campaign, coverage_stats)
        return CampaignStats(input_stats=input_stats, coverage_stats=coverage_stats, exec_stats=exec_stats,
                             seed_sharing_stats=seed_sharing_stats, smt_stats=smt_stats)


    def _calcul_triton_input_to_broker(self, campaign: CampaignResult) -> Dict[str, str]:
        mapping = {}

        def iter_input_dir(conf, dirname):
            tt_workspace = campaign.workspace.root / conf.workspace
            for file in (tt_workspace / dirname).iterdir():
                try:
                     raw = base64.b64decode(json.loads(file.read_text())['files']['input_file'])
                except:
                    raw = file.read_bytes()
                mapping[md5(raw).hexdigest()] = str(file)

        for fuzzer, items in campaign.results:
            if campaign.is_triton(fuzzer):
                conf = campaign.fuzzers_config[fuzzer]
                iter_input_dir(conf, "corpus")
                iter_input_dir(conf, "worklist")
                iter_input_dir(conf, "crashes")

        return mapping

    def _calcul_input_stats(self, campaign: CampaignResult) -> List[InputEntry]:
        entries = []

        # FIXME: Compute uniquness
        useless_ctrs = Counter()

        for fuzzer, items in campaign.results:
            num = len(items)
            for item in items:
                if not len(item.overall_new_items_covered):
                    useless_ctrs[fuzzer] += 1
            syms = {"CC": 0, "SR": 0, "SW": 0, "DYN": 0}
            if campaign.is_triton(fuzzer):
                conf = campaign.fuzzers_config[fuzzer]
                tt_workspace = campaign.workspace.root / conf.workspace
                for file in (tt_workspace / "corpus").iterdir():
                    for s in syms:
                        if s in str(file):
                            syms[s] += 1
                for file in (tt_workspace / "worklist").iterdir():
                    for s in syms:
                        if s in str(file):
                            syms[s] += 1
                for file in (tt_workspace / "crashes").iterdir():
                    for s in syms:
                        if s in str(file):
                            syms[s] += 1

            entry = InputEntry(engine=fuzzer, number=num, unique=-1, useless=useless_ctrs[fuzzer], condition=syms["CC"],
                               symread=syms["SR"], symwrite=syms["SW"], symjump=syms["DYN"])
            entries.append(entry)
        return entries

    def _calcul_coverage_stats(self, campaign: CampaignResult) -> List[CoverageEntry]:
        # all_cov = campaign.fuzzers_items[campaign.ALL_FUZZER]
        seed_cov = campaign.fuzzers_coverage[campaign.SEED_FUZZER]

        entries = []

        firsts = Counter()
        for entry in campaign.delta_items:
            firsts[entry.fuzzer] += len(entry.overall_new_items_covered)

        for fuzzer, items in campaign.results:
            cov = campaign.fuzzers_coverage[fuzzer]
            num = len(cov.difference(seed_cov)) if fuzzer != campaign.SEED_FUZZER else cov.unique_covitem_covered

            # FIXME: Compute unique & first
            first = firsts[fuzzer]
            entry = CoverageEntry(engine=fuzzer, number=num, unique=-1, first=first, total=cov.unique_covitem_covered)
            entries.append(entry)
        return entries

    def _calcul_exec_stats(self, campaign: CampaignResult) -> List[ExecEntry]:
        entries = []

        for fuzzer, config in campaign.fuzzers_config.items():
            try:
                if campaign.is_triton(fuzzer):
                    workdir = (campaign.workspace.root / "clients_ws") / Path(config.workspace).name
                    pstats = json.loads((workdir / "metadata/pastidse-stats.json").read_text())

                    # Timing stats
                    tot, replay_time = pstats["total_time"], pstats["replay_time"]
                    emu_time = pstats.get("emulation_time", 0)

                    solv_time = pstats.get("solving_time")
                    if solv_time is None:
                        sstats = json.loads((workdir / "metadata/solving_stats.json").read_text())
                        solv_time = sstats['total_solving_time']

                    dse = emu_time - replay_time
                    run_time = dse + replay_time + solv_time
                    wait = tot - run_time

                    entry = ExecEntry(engine=fuzzer, dse=dse, smt=solv_time, replay=replay_time, total=run_time, wait=wait)
                    entries.append(entry)
            except FileNotFoundError:
                logging.error(f"can't find Triton stats for {fuzzer}")
        return entries

    def _calcul_seed_sharing_stats(self, campaign: CampaignResult) -> List[SeedSharingEntry]:
        entries = []

        for fuzzer, config in campaign.fuzzers_config.items():
            try:
                if campaign.is_triton(fuzzer):
                    workdir = (campaign.workspace.root / "clients_ws") / Path(config.workspace).name
                    pstats = json.loads((workdir / "metadata/pastidse-stats.json").read_text())
                    tots, accs, rejs = pstats["seed_received"], pstats["seed_accepted"], pstats["seed_rejected"]
                    ratio = accs/tots if rejs else 1

                    entry = SeedSharingEntry(engine=fuzzer, accepted=accs, rejected=rejs, total=tots, ratio=ratio)
                    entries.append(entry)
            except FileNotFoundError:
                logging.error(f"can't find Triton stats for {fuzzer}")
        return entries


    def _calcul_smt_stats(self, campaign: CampaignResult, cov_results: List[CoverageEntry]) -> List[SmtEntry]:
        cov_data = {cov.engine: cov for cov in cov_results}

        entries = []

        for fuzzer, config in campaign.fuzzers_config.items():
            try:
                if campaign.is_triton(fuzzer):
                    cov_number = cov_data[fuzzer].number if fuzzer in cov_data else 0
                    workdir = (campaign.workspace.root / "clients_ws") / Path(config.workspace).name
                    sstats = json.loads((workdir / "metadata/solving_stats.json").read_text())

                    # Solving stats
                    sovt = sstats['total_solving_time']
                    stot, sat, unsat, to = sstats["total_solving_attempt"], sstats["SAT"], sstats["UNSAT"], sstats["TIMEOUT"]
                    coved, uncoved = len(sstats["branch_reverted"]), len(sstats["branch_not_solved"])
                    ratio = cov_number / sat if sat else cov_number
                    entry = SmtEntry(engine=fuzzer, sat=sat, unsat=unsat, timeout=to, total=stot, avg_query=sovt/stot,
                                     cov_sat_ratio=ratio, branch_solved=coved, branch_not_solved=uncoved)
                    entries.append(entry)
            except FileNotFoundError:
                logging.error(f"can't find Triton stats for {fuzzer}")
        return entries


    def print_stats(self, campaign: CampaignResult, stats: CampaignStats):
        console = Console()

        def sfmt(seconds, total, w = True) -> str:
            m, s = divmod(seconds, 60)
            h, m = divmod(m, 60)
            s = int(s) if int(s) > 0 else f"{s:.2f}"
            t = (f"{int(h)}h" if h else '') + f"{int(m)}m{s}s"
            perc = f"{seconds / total:.2%}"
            return f"{t} ({perc})" if w else t


        # InputEntry
        table = Table(show_header=True, title="INPUT", header_style="bold magenta")
        for name in ["engine", "number", "unique", "useless", "CC", "SR", "SW", "SDYN"]:
            table.add_column(name)
        for it in stats.input_stats:
            fname = self.format_fuzzer_name(campaign, it.engine, False)
            useless = f"{it.useless} ({it.useless / it.number:.2%})"
            table.add_row(fname, str(it.number), str(it.unique), useless, str(it.condition), str(it.symread), str(it.symwrite), str(it.symjump))
        console.print(table)

        # Coverage
        table = Table(show_header=True, title="COVERAGE", header_style="bold magenta")
        for name in ["engine", "edge-cov", "unique", "first", "total"]:
            table.add_column(name)
        for it in stats.coverage_stats:
            fname = self.format_fuzzer_name(campaign, it.engine, False)
            table.add_row(fname, str(it.number), str(it.unique), str(it.first), str(it.total))
        console.print(table)

        # ExecEntry
        table = Table(show_header=True, title="EXECUTION", header_style="bold magenta")
        for name in ["engine", "DSE", "SMT", "replay", "total", "wait"]:
            table.add_column(name)
        for it in stats.exec_stats:
            fname = self.format_fuzzer_name(campaign, it.engine, False)
            tot = it.total
            table.add_row(fname, sfmt(it.dse, tot), sfmt(it.smt, tot), sfmt(it.replay, tot), sfmt(tot, tot, False), sfmt(it.wait, tot, False))
        console.print(table)

        # SeedSharingEntry
        table = Table(show_header=True, title="SEED SHARING", header_style="bold magenta")
        for name in ["engine", "accepted", "rejected", "total"]:
            table.add_column(name)
        for it in stats.seed_sharing_stats:
            fname = self.format_fuzzer_name(campaign, it.engine, False)
            acc = f"{it.accepted} ({it.accepted / it.total:.2%})"
            rejs = f"{it.rejected} ({it.rejected / it.total:.2%})"
            table.add_row(fname, acc, rejs, str(it.total))
        console.print(table)

        # SmtEntry
        table = Table(show_header=True, title="SMT SOLVING", header_style="bold magenta")
        for name in ["engine", "SAT", "UNSAT", "TO", "Total", "mean query", "cov/input", "branches solved", "branches not solved"]:
            table.add_column(name)
        for it in stats.smt_stats:
            fname = self.format_fuzzer_name(campaign, it.engine, False)
            table.add_row(fname, str(it.sat), str(it.unsat), str(it.timeout), str(it.total), f"{it.avg_query:.2f}", f"{it.cov_sat_ratio:.2f}", str(it.branch_solved), str(it.branch_not_solved))
        console.print(table)


        # for stat in (getattr(stats, x) for x in stats.schema()['properties']):
        #     if not stat:
        #         print(f"Stat {stat} is empty")
        #         continue
        #     table = Table(show_header=True, title=str(type(stat[0])), header_style="bold magenta")
        #     item = stat[0]
        #
        #     for name, column in {x: getattr(item, x) for x in item.schema()['properties']}.items():
        #         table.add_column(name)
        #     for item in stat:
        #         table.add_row(*[str(getattr(item, x)) for x in item.schema()['properties']])
        #     console.print(table)



    def show_delta_history(self, campaign: CampaignResult) -> None:
        console = Console()
        def tt(secs):
            return str(timedelta(seconds=int(secs)))
        def pp_edge(e):
            return f"({e[0]:#08x}-{e[1]:#08x})"

        mapping = self._calcul_triton_input_to_broker(campaign)

        def caract_input(name) -> str:
            types = ["_CC_", "_SR_", "_SW_", "_DYN_"]
            for t in types:
                if t in name:
                    return t[1:-1]
            return "N/C"

        table = Table(show_header=True, title="Delta History", header_style="bold magenta")
        for col in ["Elapsed", "Fuzzer", "Type", "New", "Tot Cov", "Items"]:
            table.add_column(col)
        # FIXME: Ajouter une colonne pour le count

        for delta in campaign.delta_items:

            # Resolve broker input to triton ones (to what it was generated from)
            typ = "-"
            if campaign.is_triton(delta.fuzzer):
                meta = campaign.parse_filename(delta.input_name)
                if meta:
                    hash = meta[3]
                    triton_input_name = mapping.get(hash)
                    if triton_input_name:
                        typ = caract_input(triton_input_name)

            table.add_row(tt(delta.time_elapsed),
                          self.format_fuzzer_name(campaign, delta.fuzzer, short=True),
                          typ,
                          str(len(delta.overall_new_items_covered)),
                          str(delta.overall_coverage_sum),
                          " ".join(pp_edge(e) for e in list(delta.overall_new_items_covered)[:4])
            )

        console.print(table)



# built-in imports
from pathlib import Path
from typing import Union, Tuple, List, Optional
import json

# third-party import
import lief

from libpastis import FuzzingEngineDescriptor, EngineConfiguration
from libpastis.types import ExecMode, CoverageMode, FuzzMode

# WARNING: This module is made in such a way that it does
# not directly depend on tritondse (to facilitate installation)
# coverage strategies thus have to be ported here !
TRITON_DSE_COVS = ['BLOCK', 'EDGE', 'PATH']


class TritonConfigurationInterface(EngineConfiguration):
    def __init__(self, data):
        self.data = data

    @staticmethod
    def new() -> 'TritonConfigurationInterface':
        return TritonConfigurationInterface({})

    @staticmethod
    def from_file(filepath: Path) -> 'TritonConfigurationInterface':
        with open(filepath, "r") as f:
            return TritonConfigurationInterface(json.load(f))

    @staticmethod
    def from_str(s: str) -> 'TritonConfigurationInterface':
        return TritonConfigurationInterface(json.loads(s))

    def to_str(self) -> str:
        return json.dumps(self.data)

    def get_coverage_mode(self) -> CoverageMode:
        """ Current coverage mode selected in the file """
        v = self.data['coverage_strategy']
        return CoverageMode(v)

    def set_target(self, target: int) -> None:
        self.data['custom'] = {}
        self.data['custom']['target'] = target


class TritonEngineDescriptor(FuzzingEngineDescriptor):

    NAME = "TRITON"
    SHORT_NAME = "TT"
    VERSION = "1.0.0"

    FUNCTION_BLACKLIST_PREFIX = [
        "__sanitizer",  # all fuzzer related sanitizers
        "__gcov_",      # gcov functions
        "__asan_",
        "__afl_"
    ]

    SYMBOL_BLACKLIST_PREFIX = [
        "__sanitizer",  # all fuzzer related sanitizers
        "__gcov_",      # gcov functions
        "__asan_",
        "__afl_"
    ]

    IMPORT_BLACKLIST = [
        "HF_ITER"       # honggfuzz functions
    ]

    config_class = TritonConfigurationInterface

    def __init__(self):
        pass

    @staticmethod
    def accept_file(binary_file: Path) -> Tuple[bool, Optional[ExecMode], Optional[FuzzMode]]:
        p = lief.parse(str(binary_file))
        if not p:
            return False, None, None

        # Presumably good unless do find some instrumentation functions or imports
        for f in p.functions:
            for item in TritonEngineDescriptor.FUNCTION_BLACKLIST_PREFIX:
                if f.name.startswith(item):
                    return False, None, None

        for f in p.imported_functions:
            if f.name in TritonEngineDescriptor.IMPORT_BLACKLIST:
                return False, None, None

        for s in p.symbols:
            for item in TritonEngineDescriptor.SYMBOL_BLACKLIST_PREFIX:
                if s.name.startswith(item):
                    return False, None, None

        return True, ExecMode.SINGLE_EXEC, FuzzMode.BINARY_ONLY  # Only support single_exec, binary only (not instrumented)

    @staticmethod
    def supported_coverage_strategies() -> List[CoverageMode]:
        return [CoverageMode(st) for st in TRITON_DSE_COVS]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: message.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rmessage.proto\x12\tlibpastis\"@\n\rFuzzingEngine\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x10\n\x08pymodule\x18\x03 \x01(\t\"x\n\x0cInputSeedMsg\x12\x0c\n\x04seed\x18\x01 \x01(\x0c\x12.\n\x04type\x18\x02 \x01(\x0e\x32 .libpastis.InputSeedMsg.SeedType\"*\n\x08SeedType\x12\t\n\x05INPUT\x10\x00\x12\t\n\x05\x43RASH\x10\x01\x12\x08\n\x04HANG\x10\x02\"\x17\n\x07\x44\x61taMsg\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\t\"\xde\x04\n\x08StartMsg\x12\x17\n\x0f\x62inary_filename\x18\x01 \x01(\t\x12\x0e\n\x06\x62inary\x18\x02 \x01(\x0c\x12\x13\n\x0bsast_report\x18\x03 \x01(\x0c\x12(\n\x06\x65ngine\x18\x04 \x01(\x0b\x32\x18.libpastis.FuzzingEngine\x12/\n\texec_mode\x18\x05 \x01(\x0e\x32\x1c.libpastis.StartMsg.ExecMode\x12/\n\tfuzz_mode\x18\x06 \x01(\x0e\x32\x1c.libpastis.StartMsg.FuzzMode\x12\x31\n\ncheck_mode\x18\x07 \x01(\x0e\x32\x1d.libpastis.StartMsg.CheckMode\x12\x15\n\rcoverage_mode\x18\x08 \x01(\t\x12\x38\n\rseed_location\x18\t \x01(\x0e\x32!.libpastis.StartMsg.SeedInjectLoc\x12\x13\n\x0b\x65ngine_args\x18\n \x01(\t\x12\x14\n\x0cprogram_argv\x18\x0b \x03(\t\":\n\x08\x45xecMode\x12\r\n\tAUTO_EXEC\x10\x00\x12\x0f\n\x0bSINGLE_EXEC\x10\x01\x12\x0e\n\nPERSISTENT\x10\x02\"<\n\x08\x46uzzMode\x12\r\n\tAUTO_FUZZ\x10\x00\x12\x10\n\x0cINSTRUMENTED\x10\x01\x12\x0f\n\x0b\x42INARY_ONLY\x10\x02\"9\n\tCheckMode\x12\r\n\tCHECK_ALL\x10\x00\x12\x0e\n\nALERT_ONLY\x10\x01\x12\r\n\tALERT_ONE\x10\x02\"$\n\rSeedInjectLoc\x12\t\n\x05STDIN\x10\x00\x12\x08\n\x04\x41RGV\x10\x01\"\t\n\x07StopMsg\"\xf1\x01\n\x08HelloMsg\x12.\n\x0c\x61rchitecture\x18\x01 \x01(\x0e\x32\x18.libpastis.HelloMsg.Arch\x12\x0c\n\x04\x63pus\x18\x02 \x01(\r\x12\x0e\n\x06memory\x18\x03 \x01(\x04\x12)\n\x07\x65ngines\x18\x04 \x03(\x0b\x32\x18.libpastis.FuzzingEngine\x12\x10\n\x08hostname\x18\x05 \x01(\t\x12%\n\x08platform\x18\x06 \x01(\x0e\x32\x13.libpastis.Platform\"3\n\x04\x41rch\x12\x07\n\x03X86\x10\x00\x12\n\n\x06X86_64\x10\x01\x12\t\n\x05\x41RMV7\x10\x02\x12\x0b\n\x07\x41\x41RCH64\x10\x03\"\x8b\x01\n\x06LogMsg\x12\x0f\n\x07message\x18\x01 \x01(\t\x12)\n\x05level\x18\x02 \x01(\x0e\x32\x1a.libpastis.LogMsg.LogLevel\"E\n\x08LogLevel\x12\t\n\x05\x44\x45\x42UG\x10\x00\x12\x08\n\x04INFO\x10\x01\x12\x0b\n\x07WARNING\x10\x02\x12\t\n\x05\x45RROR\x10\x03\x12\x0c\n\x08\x43RITICAL\x10\x04\"\xfe\x01\n\x0cTelemetryMsg\x12\x1f\n\x05state\x18\x01 \x01(\x0e\x32\x10.libpastis.State\x12\x14\n\x0c\x65xec_per_sec\x18\x02 \x01(\r\x12\x12\n\ntotal_exec\x18\x03 \x01(\x04\x12\r\n\x05\x63ycle\x18\x04 \x01(\r\x12\x0f\n\x07timeout\x18\x05 \x01(\r\x12\x16\n\x0e\x63overage_block\x18\x06 \x01(\r\x12\x15\n\rcoverage_edge\x18\x07 \x01(\r\x12\x15\n\rcoverage_path\x18\x08 \x01(\r\x12\x17\n\x0flast_cov_update\x18\t \x01(\x04\x12\x11\n\tcpu_usage\x18\n \x01(\x02\x12\x11\n\tmem_usage\x18\x0b \x01(\x02\"\x16\n\x14StopCoverageCriteria\"\xf8\x02\n\x0b\x45nvelopeMsg\x12,\n\tinput_msg\x18\x01 \x01(\x0b\x32\x17.libpastis.InputSeedMsgH\x00\x12&\n\x08\x64\x61ta_msg\x18\x02 \x01(\x0b\x32\x12.libpastis.DataMsgH\x00\x12(\n\tstart_msg\x18\x03 \x01(\x0b\x32\x13.libpastis.StartMsgH\x00\x12&\n\x08stop_msg\x18\x04 \x01(\x0b\x32\x12.libpastis.StopMsgH\x00\x12(\n\thello_msg\x18\x05 \x01(\x0b\x32\x13.libpastis.HelloMsgH\x00\x12$\n\x07log_msg\x18\x06 \x01(\x0b\x32\x11.libpastis.LogMsgH\x00\x12\x30\n\rtelemetry_msg\x18\x07 \x01(\x0b\x32\x17.libpastis.TelemetryMsgH\x00\x12\x38\n\rstop_crit_msg\x18\x08 \x01(\x0b\x32\x1f.libpastis.StopCoverageCriteriaH\x00\x42\x05\n\x03msg*\x1e\n\x05State\x12\x0b\n\x07RUNNING\x10\x00\x12\x08\n\x04IDLE\x10\x01*L\n\x08Platform\x12\x07\n\x03\x41NY\x10\x00\x12\t\n\x05LINUX\x10\x01\x12\x0b\n\x07WINDOWS\x10\x02\x12\t\n\x05MACOS\x10\x03\x12\x0b\n\x07\x41NDROID\x10\x04\x12\x07\n\x03IOS\x10\x05\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'message_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _STATE._serialized_start=1907
  _STATE._serialized_end=1937
  _PLATFORM._serialized_start=1939
  _PLATFORM._serialized_end=2015
  _FUZZINGENGINE._serialized_start=28
  _FUZZINGENGINE._serialized_end=92
  _INPUTSEEDMSG._serialized_start=94
  _INPUTSEEDMSG._serialized_end=214
  _INPUTSEEDMSG_SEEDTYPE._serialized_start=172
  _INPUTSEEDMSG_SEEDTYPE._serialized_end=214
  _DATAMSG._serialized_start=216
  _DATAMSG._serialized_end=239
  _STARTMSG._serialized_start=242
  _STARTMSG._serialized_end=848
  _STARTMSG_EXECMODE._serialized_start=631
  _STARTMSG_EXECMODE._serialized_end=689
  _STARTMSG_FUZZMODE._serialized_start=691
  _STARTMSG_FUZZMODE._serialized_end=751
  _STARTMSG_CHECKMODE._serialized_start=753
  _STARTMSG_CHECKMODE._serialized_end=810
  _STARTMSG_SEEDINJECTLOC._serialized_start=812
  _STARTMSG_SEEDINJECTLOC._serialized_end=848
  _STOPMSG._serialized_start=850
  _STOPMSG._serialized_end=859
  _HELLOMSG._serialized_start=862
  _HELLOMSG._serialized_end=1103
  _HELLOMSG_ARCH._serialized_start=1052
  _HELLOMSG_ARCH._serialized_end=1103
  _LOGMSG._serialized_start=1106
  _LOGMSG._serialized_end=1245
  _LOGMSG_LOGLEVEL._serialized_start=1176
  _LOGMSG_LOGLEVEL._serialized_end=1245
  _TELEMETRYMSG._serialized_start=1248
  _TELEMETRYMSG._serialized_end=1502
  _STOPCOVERAGECRITERIA._serialized_start=1504
  _STOPCOVERAGECRITERIA._serialized_end=1526
  _ENVELOPEMSG._serialized_start=1529
  _ENVELOPEMSG._serialized_end=1905
# @@protoc_insertion_point(module_scope)



#built-in imports
import logging
from typing import Optional
import importlib
import inspect

from libpastis.enginedesc import FuzzingEngineDescriptor


HF_PERSISTENT_SIG = b"\x01_LIBHFUZZ_PERSISTENT_BINARY_SIGNATURE_\x02\xFF"


def load_engine_descriptor(py_module: str) -> Optional[FuzzingEngineDescriptor]:
    try:
        mod = importlib.import_module(py_module)
        mems = inspect.getmembers(mod, lambda x: inspect.isclass(x) and issubclass(x, FuzzingEngineDescriptor) and x != FuzzingEngineDescriptor)
        if not mems:
            logging.error(f"can't find FuzzingEngineDescriptor in module {py_module}")
            return None
        else:
            if len(mems) > 1:
                logging.warning(f"module {py_module} contain multiple subclass of {FuzzingEngineDescriptor} (take first)")
            return mems[0][1]
    except ImportError:
        logging.error(f"cannot import py_module: {py_module}")



COLORS = [32, 33, 34, 35, 36, 37, 39, 91, 93, 94, 95, 96]


class Bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def mk_color(text: str, color: str) -> str:
    return color+text+Bcolors.ENDC
